!*****************************************************************************************
!MODULE SAMPLER-------------------------------------------------------- BY SML. 05/07/2012
!*****************************************************************************************

 MODULE SAMPLER

!*****************************************************************************************
!CONTAINS SUBROUTINES AND FUNCTIONS USED TO SAMPLE FROM SEVERAL PROBABILITY DISTRIBUTIONS,
!INCLUDING STANDARIZED UNIFORM (RAN), STANDARIZED NORMAL (GASDEV), GAMMA (RGAMMA), CENTRAL
!CHI-SQUARE (CHSQ), SCALED INVERTED CHI-SQUARE (ICHSQ) AND WISHART (WSHRT).
!*****************************************************************************************
!
!VERSION 2.0******************************************************************************
!UPDATE OBSOLETE STRUCTURES FROM THE F77 SAMPLING ROUTINES AND INTRODUCE NOVEL ONES.
!*****************************************************************************************
!
!ATT. AUXILIARY MODULE "PARAM" WITH NUMERICAL PRECISION DEFINITION AND SEED IS REQUIRED.
!***
!
!-------
!MODULES
!-------
 USE PARAM
!---
!
 CONTAINS
!
!*****************************************************************************************
!FUNCTION RAN------------------------------------------------------------------ 26/06/2012
!*****************************************************************************************
!
!RETURNS A UNIFORM RANDOM DEVIATE BETWEEN 0.0 AND 1.0 (EXCLUSIVE OF THE ENDPOINT VALUES)
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!"MINIMAL" RANDOM NUMBER GENERATOR OF PARK & MILLER COMBINED WITH A MARSAGLIA SHIFT SEQUEN-
!CE. THIS FULLY  PORTABLE, SCALAR GENERATOR HAS THE "TRADITIONAL" (NOT FORTRAN 90) CALLING
!SEQUENCE WITH A RANDOM DEVIATE AS THE  RETURNED FUNCTION VALUE: CALL WITH "IDUM" SET TO A
!NEGATIVE INTEGER TO INITIALIZE; THEREAFTER, DO NOT ALTER IDUM EXCEPT TO REINITIALIZE. THE
!PERIOD OF THIS GENERATOR IS ABOUT 3.1E18. (CODE HAND-COPIED FROM PRESS ET AL. BOOK: "NUME-
!MERICAL RECIPES IN FORTRAN 90")
!*****************************************************************************************

 FUNCTION RAN(IDUM)

!***********
!DECLARATION
!***********
 IMPLICIT NONE

!DUMMY ARGUMENT: SEED
 INTEGER(K4B), INTENT(INOUT):: IDUM

!LOCAL VARIABLES
 INTEGER(K4B), PARAMETER:: IA=16807, IM=2147483647, IQ=127773, IR=2836
 REAL(DP), SAVE:: AM

 INTEGER(K4B), SAVE:: IX=-1, IY=-1, K

 REAL(DP):: RAN
!***

!*********
!PROCEDURE
!*********
      
!INITIALIZATION OR REINITIALIZATION 
 IF(IDUM <= 0 .OR. IY < 0) THEN

	 AM=NEAREST(1.0,-1.0)/IM
	 IY=IOR(IEOR(888889999,ABS(IDUM)),1)
	 IX=IEOR(777755555,ABS(IDUM))
	 IDUM=ABS(IDUM)+1 !SET IDUM POSITIVE
 
 ENDIF

!MARSAGLIA SHIFT SEQUENCE WITH PERIOD 2^32 - 1
 IX=IEOR(IX,ISHFT(IX,13))
 IX=IEOR(IX,ISHFT(IX,-17))
 IX=IEOR(IX,ISHFT(IX,5))

!PARK-MILLER SEQUENCE BY SCHRAGE'S METHOD, PERIOD 2^32 - 2
 K=IY/IQ
 IY=IA*(IY-K*IQ)-IR*K
 IF(IY<0) IY=IY+IM

!COMBINE THE TWO GENERATORS WITH MASKING TO ENSURE NONZERO VALUE
 RAN=AM*IOR(IAND(IM,IEOR(IX,IY)),1) 

 END FUNCTION RAN
!***
!
!*****************************************************************************************
!FUNCTION GASDEV--------------------------------------------------------------- 28/06/2012
!*****************************************************************************************
!
!RETURNS A STANDARIZED NORMAL RANDOM VARIABLE REALIZATION.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!BOX-MULLER METHOD WITH A SMALL TRICK TO AVOID REFERENCING TRIGONOMETRIC FUNTIONS. THE ME-
!THOD GENERATES TWO INDEPENDENT NORMAL DEVIATES, BUT RETURNS A SINGLE ONE IN EACH CALL. IT
!STORES THE OTHER DEVIATE FOR A SUBSEQUENT CALL. SEE THE BOOK BY PRESS ET AL. "NUMEMERICAL
!RECIPES" FOR FURTHER DETAILS.
!*****************************************************************************************
!
!ATT. AUXILIARY FUNCTION "RAN" IS REQUIRED.
!***
!
 FUNCTION GASDEV(IDUM)

!***********
!DECLARATION
!***********
 IMPLICIT NONE

!DUMMY ARGUMENT: SEED
 INTEGER(K4B), INTENT(INOUT):: IDUM

!LOCAL VARIABLES
 LOGICAL, SAVE:: ISET=.FALSE.
 REAL(KIND=DP):: GASDEV, RSQ, V1, V2
 REAL(KIND=DP), SAVE:: GSET

!NOTE-------------------------------------------------------------------------------------
!"SAVE" IS A SPECIFICATION STATEMENT WHICH CAN BE USED TO ENSURE THAT VARIABLES AND ARRAYS
!USED WITHIN A PROCEDURE PRESERVE THEIR VALUES BETWEEN SUCCESSIVE CALLS TO THE PROCEDURE.
!-----------------------------------------------------------------------------------------
!***

!*********
!PROCEDURE
!*********

!ISET FLAGS IF THE 2ND DEVIATE HAS ALREADY BEEN USED
 IF (ISET) THEN

	 GASDEV = GSET
	 ISET = .FALSE.	!RESET FLAG

 ELSE

	 DO

		!PICK TWO UNIFORM NUMBERS IN THE INTERVAL (-1, +1)
		 V1 = 2.D0*RAN(IDUM)-1.D0
		 V2 = 2.D0*RAN(IDUM)-1.D0

		!COMPUTE THEIR SUM OF SQUARES
	     RSQ = V1**2 + V2**2
     
		!CHECK IF THE NUMBER IS IN THE UNIT CIRCLE
		 IF(RSQ < 1.D0 .AND. RSQ > 0.D0) EXIT

	 ENDDO

	!COMPUTE DEVIATES
	 RSQ = DSQRT(-2.D0*DLOG(RSQ)/RSQ)
	 GASDEV = V1*RSQ
     GSET = V2*RSQ

	!SET FLAG
	 ISET = .TRUE.

 ENDIF
 
 END FUNCTION GASDEV
!***
!
!*****************************************************************************************
!RECURSIVE FUNCTION RGAMMA ---------------------------------------------------- 04/07/2012
!*****************************************************************************************
!
!RETURNS A RANDOM DEVIATE FROM A GAMMA(SHAPE, SCALE) DISTRIBUTION.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!IMPLEMENTATION BASED ON "A SIMPLE METHOD FOR GENERATING GAMMA VARIABLES" BY GEORGE MARSA-
!GLIA AND WAI WAN TSANG (2000). ACM TRANSACTIONS ON MATHEMATICAL SOFTWARE, 26(3): 363-372.
!CODE ADAPTED FROM SUKHBINDER's PROJECTS WEBSITE.
!*****************************************************************************************
!
!ATT. AUXILIARY FUNCTIONS "RAN" & "GASDEV" ARE REQUIRED.
!***
!
 RECURSIVE FUNCTION RGAMMA(SHAPE, SCALE, IDUM) RESULT(ANS)
!
!***********
!DECLARATION
!***********
 IMPLICIT NONE
 
!DUMMY ARGUMENTS:
 REAL(DP), INTENT(IN):: SHAPE, SCALE
 INTEGER(K4B), INTENT(INOUT):: IDUM

!LOCAL VARIABLES:
 REAL(DP):: U, V, W, D, C, X, XSQ, G, ANS
!***

!*********
!PROCEDURE
!*********
 IF (SHAPE <= 0.0D0) THEN
	 WRITE(*,*) "SHAPE PARAMETER MUST BE POSITIVE"
 END IF

 IF (SCALE <= 0.0D0) THEN
	 WRITE(*,*) "SCALE PARAMETER MUST BE POSITIVE"
 END IF
!---

 IF (SHAPE >= 1.0D0) THEN
	 D = SHAPE - 1.0D0/3.0D0
	 C = 1.0D0/(9.0D0*D)**0.5D0

     DO WHILE (.TRUE.)

		 X = GASDEV(IDUM)
		 V = 1.0D0 + C*X

		 DO WHILE (v <= 0.0D0)
			 X = GASDEV(IDUM)
			 V = 1.0D0 + C*X
		 ENDDO

         V = V*V*V
		 U = RAN(IDUM)
         XSQ = X*X

		 IF (U < 1.0D0 -.0331D0*XSQ*XSQ) THEN
			 ANS = SCALE*D*V
             RETURN
		 END IF

		 IF (DLOG(U) < 0.5D0*XSQ + D*(1.0D0 - V + DLOG(V))) THEN 		 
			 ANS = SCALE*D*V
             RETURN
		 END IF

	 ENDDO

 ELSE

	 G = RGAMMA(SHAPE+1.0D0, 1.0D0, IDUM)
	 W = RAN(IDUM)
     ANS = SCALE*G*(W)**(1.0D0/SHAPE)
     RETURN

 END IF
!---

 END FUNCTION RGAMMA
!***
!
!*****************************************************************************************
!FUNCTION CHSQ ---------------------------------------------------------------- 05/07/2012
!*****************************************************************************************
!
!RETURNS A RANDOM DEVIATE FROM A CHI-SQUARE (DF) DISTRIBUTION.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!GAMMA DISTRIBUTION WITH CORRESPONDING PARAMETERS SET TO SHAPE = 0.5D0*DF AND SCALE = 2.D0
!*****************************************************************************************
!
!ATT. AUXILIARY FUNCTION "RGAMMA" IS REQUIRED.
!***
!
 FUNCTION CHSQ(DF, IDUM) RESULT(ANS)
!
!***********
!DECLARATION
!***********
 IMPLICIT NONE
 
!DUMMY ARGUMENTS:
 REAL(DP), INTENT(IN):: DF
 INTEGER(K4B), INTENT(INOUT):: IDUM

!LOCAL VARIABLES:
 REAL(DP):: ANS

!*********
!PROCEDURE
!*********

 ANS=RGAMMA(0.5D0*DF, 2.0D0, IDUM)

 END FUNCTION CHSQ
!***
!
!*****************************************************************************************
!FUNCTION ICHSQ --------------------------------------------------------------- 05/07/2012
!*****************************************************************************************
!
!RETURNS A SCALED INVERTED CHI-SQUARE RANDOM VARIABLE REALIZATION.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!FIRST, AN INVERSE CHI-SQUARE DEVIATE IS SAMPLED USING CHSQ(DF). THEN THE RESULT IS SCALED
!TO "SCALE" PARAMETER. 
!*****************************************************************************************
!
!ATT. AUXILIARY FUNCTION "CHSQ" IS REQUIRED.
!***
!
 FUNCTION ICHSQ(DF, SCALE, IDUM) RESULT(ANS)
!
!***********
!DECLARATION
!***********
 IMPLICIT NONE
 
!DUMMY ARGUMENTS:
 REAL(DP), INTENT(IN):: DF, SCALE
 INTEGER(K4B), INTENT(INOUT):: IDUM

!LOCAL VARIABLES:
 REAL(DP):: ANS

!*********
!PROCEDURE
!*********
 ANS=CHSQ(DF, IDUM)
 ANS=SCALE/ANS

 END FUNCTION ICHSQ
!***
!
!*****************************************************************************************
!SUBROUTINE WSHRT ----------------------------------------------------- BY SML. 18/06/2009
!*****************************************************************************************
!
!RETURNS A WISHART RANDOM VARIABLE REALIZATION. ON OUTPUT, "SA" IS AN UPPER TRIANGULAR MA-
!TRIX OF SIZE NPxNP (WRITTEN IN LINEAR FORM, COLUMN ORDERED) WHOSE ELEMENTS HAVE A WISHART
!(N, SIGMA) DISTRIBUTION. D IS AN UPPER TRIANGULAR ARRAY SUCH THAT SIGMA = D'D. IT MUST BE
!SUPPLIED IN LINEAR FORM, COLUMN ORDERED. THE ORIGINAL SUBROUTINE GENERATES THE SAMPLE CO-
!VARIANCE MATRIX OF AN HYPOTHETICAL RANDOM SAMPLE OF SIZE N, EACH DISTRIBUTED AS A NP-NOR-
!MAL. SO IT HAS BEEN SLIGHTLY CHANGED FROM THE ORIGINAL CODE, ASIDE THE F90 ADAPTATION, TO
!RETURN ONLY THE WISHART MATRIX.
!***
!
!ATT. AUXILIARY FUNCTIONS "GASDEV" AND "CHSQ" ARE REQUIRED.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!SMITH & HOCKING (1972), APPLIED STATISTICS, VOL.21, N3: ALGORITHM AS 53: WISHART VARIATE
!GENERATOR. IT'S BASED IN SETTING UP AN UPPER TRIANGULAR (NPxNP)MATRIX "SB", SUCH THAT ITS
!DIAGONAL ENTRIES ARE SAMPLED FROM SQUARE ROOTED CHI-SQUARE VARIABLES AND ITS OFF-DIAGONAL
!ARE TAKEN FROM STANDARIZED NORMAL DEVIATES. LET THE SUBROUTINE'S ARGUMENT "D" BE AN UPPER
!TRIANGULAR MATRIX SUCH THAT D'D = SCALE MATRIX. THEN, THE MATRIX SA = D'(SB)'(SB)D IS DIS-
!TRIBUTED AS A WISHART(N, SCALE).
!*****************************************************************************************

 SUBROUTINE WSHRT(D, N, NP, NNP, SA, IDUM)

!INTENT(IN)::     D - UPPER TRIANGULAR MATRIX, STORED COLUMN ORDERED AS A VECTOR, SUCH
!			          THAT D'D = SCALE PARAMETER.
!INTENT(IN)::     N - DEGREES OF BELIEF PARAMETER.
!INTENT(IN)::    NP - DIMENSION OF THE MATRIX.
!INTENT(IN)::   NNP - DIMENSION OF D VECTOR, SUCH THAT NNP = NP*(NP+1)/2.
!INTENT(IN)::  IDUM - SEED FOR THE RANDOM NUMBER GENERATOR.
!INTENT(INOUT):: SA - SAMPLE FROM A WISHART(N, SCALE) DENSITY, WRITTEN IN LINEAR FORM.

!***********
!DECLARATION
!***********
 IMPLICIT NONE

!DUMMY ARGUMENTS
 INTEGER(K4B), INTENT(INOUT):: IDUM !SEED
 INTEGER(K4B), INTENT(IN):: NP, NNP

 REAL(KIND=DP), INTENT(IN):: N, D(NNP)
 REAL(KIND=DP), INTENT(INOUT):: SA(NNP)

!LOCAL VARIABLES 
 INTEGER(K4B):: NS, NR, IP, NQ
 REAL(KIND=DP):: SB(NNP)
 REAL(KIND=DP):: DF, U1, U2, RN, C
 REAL(KIND=DP):: ZERO = 0.D0
 INTEGER(K4B):: k, i, j, ii

!*********
!PROCEDURE
!*********

 k = 1

 DO

	!RETRIEVE TWO INDEPENDENT RANDOM NORMAL DEVIATES
	 U1 = GASDEV(IDUM)
	 U2 = GASDEV(IDUM)

	 !LOAD "SB" WITH INDEPENDENT NORMAL(0,1) VARIATES
	  SB(k) = U1
	  k = k + 1
	  IF (k .GT. NNP) EXIT
	  SB(k) = U2
	  k = k + 1
	  IF (k .LE. NNP) CYCLE

	  EXIT

 ENDDO

 NS = 0

!LOAD DIAGONAL ELEMENTS WITH SQUARE ROOT OF CHI-SQUARES VARIATES
 DO i = 1, NP

	 DF = N - i + 1           
	 NS = NS + i
     
	 SB(NS) = CHSQ(DF, IDUM)
	 SB(NS) = DSQRT(SB(NS))

 ENDDO

 RN = N
 NR = 1

!PERFORM D'(SB)'=(SB*D)'
 DO i = 1, NP 
 
	 NR = NR + i - 1 

	 DO j = i, NP
		
		 IP = NR
         NQ = (j*j - j) / 2 + i - 1
	     C = ZERO
         
		 DO k = i, j	 

			 IP = IP + k - 1 
	         NQ = NQ + 1 
	          C = C + SB(IP) * D(NQ)         

		 ENDDO

		 SA(IP) = C 

	 ENDDO

 ENDDO

!PERFORM (SB*D)'(SB*D)
 DO i = 1, NP

	 ii = NP - i + 1
     NQ = NNP - NP
	 
	 DO j = 1, i

		 IP = (ii*ii - ii) / 2
		  C = ZERO

		 DO k = i, NP
			 IP = IP + 1
			 NQ = NQ + 1
			  C = C + SA(IP) * SA(NQ)

         ENDDO

		!NOTE: IN THE ORIGINAL CODE, SA(:)'S ENTRIES WERE STANDARIZED
		!	   BY THE DEGREES OF BELIEF. THAT IS: SA(NQ) = C / RN.
		 SA(NQ) = C !/ RN
		 NQ = NQ - 2 * NP + i + j - 1

	 ENDDO

 ENDDO
!***

 RETURN

 END SUBROUTINE WSHRT
!***
!
 END MODULE SAMPLER
!***
