!*****************************************************************************************
!MODULE MALG----------------------------------------------------------- BY SML. 07/08/2012
!*****************************************************************************************
!
 MODULE MALG
!
!*****************************************************************************************
!CONTAINS ROUTINES USED TO PERFORM MATRIX ALGEBRA OPERATIONS.
!*****************************************************************************************
!
!***VERSION 1.0***
!
!ATT. AUXILIARY MODULE "PARAM" WITH NUMERICAL PRECISION DEFINITION IS REQUIRED.
!***
!
!-------
!MODULES
!-------
 USE PARAM
!---
!
 CONTAINS
!
!*****************************************************************************************
!SUBROUTINE GINV------------------------------------------------------- BY SML. 17/02/2010
!*****************************************************************************************
!
!RETURNS GENERALIZED INVERSE AND RANK OF THE LEADING POSITIVE SEMIDEFINITE SYMMETRIC X[N,N]
!SUBMATRIX TAKEN FROM X(M,M). OPTIONALLY, IT RETURNS THE DETERMINANT.
!***
!
!***VERSION 1.1***
!ADAPTED FROM VERSION 1.0 BY CHANGING GIX AND W RANK. MISTZAL'S ADAPTATION OF FERNANDO'S F77
!SUBROUTINE.
!
!ATTENTION!
!SUBROUTINE FAILED IF X(N,N) IS NOT POSITIVE SEMIDEFINITE. THIS CONDITION SHOULD BE CHECKED.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!THIS SUBROUTINE USES CHOLESKY DECOMPOSITION METHOD FOR A GENERALIZED INVERSE (SEBER, 1977,
!SECTION 11.5.1). IF X IS A  POSITIVE SEMIDEFINITE MATRIX, STILL HOLDS X=LL', WHERE L IS A
!LOWER TRIANGULAR MATRIX, AND ITS  ELEMENTS CAN BE COMPUTED  USING CHOLESKY ALGORITHM. HOW-
!EVER, AS N-RANK ELEMENTS WILL EQUAL ZERO (LINEAR DEPENDENCIES), THE OTHER ELEMENTS IN THE
!CORRESPONDING COLUMN ARE SET TO ZERO. A GENERALIZED INVERSE IS FINALLY COMPUTED AS: GIX =
!T'T, WHERE T=INV-L.
!*****************************************************************************************

 SUBROUTINE GINV (GIX, N, M, TOL, RANK, D)

!INTENT(IN):   N - LEADING SUBMATRIX DIMENSION.
!INTENT(IN):   M - MATRIX DIMENSION.
!INTENT(IN): TOL - ROUNDING ERROR TOLERANCE. REPRESENTS THE WORKING "ZERO": VALUES BELOW
!			       TOL, WILL BE TAKEN AS ZEROS.
!
!INTENT(INOUT):  GIX - INPUT: TARGET MATRIX. OUTPUT: IT'S GENERALIZED INVERSE.
!INTENT(INOUT): RANK - MATRIX RANK.
!
!INTENT(OUT): D - MATRIX DETERMINANT.
!
!********************
!VARIABLE DECLARATION
!********************
 IMPLICIT NONE

 INTEGER(K4B),INTENT(IN):: N, M
 INTEGER(K4B),INTENT(INOUT):: RANK
 REAL(DP),INTENT(IN):: TOL
 REAL(DP),INTENT(INOUT):: GIX(M,M)
 REAL(DP),INTENT(INOUT), OPTIONAL:: D

 LOGICAL:: CHECK !.TRUE. IF DETERMINANT IS REQUIRED.
 REAL(DP):: W(M) !WORKING VECTOR. 
 INTEGER(K4B):: i, j !DUMMY VARIABLES.
!***
!
!*********
!PROCEDURE
!*********
 RANK = N

!STEP 1. PERFORM CHOLESKY DECOMPOSITION. OBTAIN LOWER TRIANGULAR MATRIX L, SUCH THAT LL'=X.
!OVERWRITES BELOW DIAGONAL ELEMENTS OF X.
 DO i = 1, N

	 DO j = 1, i-1 
		 GIX(i:N,i) = GIX(i:N,i) - GIX(i,j)*GIX(i:N,j)
	 ENDDO
     
	!iTH DIAGONAL ENTRY IS ZERO...
	 IF (GIX(i,i).LT.TOL) THEN

		 GIX(i:N,i)	= 0.D0
		 
		 RANK = RANK-1 !FOUND A LINEAR DEPENDENCY.

	!iTH DIAGONAL ENTRY IS NOT ZERO...	 
	 ELSE

 		 GIX(i,i) = DSQRT(GIX(i,i)) !L(1,1)=SQRT[GIX(1,1)], WHEN i=1.
								    !L(i,i)=SQRT{GIX(i,i)-SUM(j=1,i-1)[L(i,j)^2]}, WHEN i=2,N.

		 GIX(i+1:N,i) = GIX(i+1:N,i)/GIX(i,i) !L(k,1)=GIX(k,1)/L(1,1), WHEN i=1 & k=2,N
		 									  !L(k,i)={GIX(k,i)-SUM(j=1,i-1)[L(i,j)*L(k,j)]}/L(i,i),
  										      !WHEN i=2,N & k=i+1,N
	 ENDIF

 ENDDO	 
!***		 

!STEP 2. IF REQUIRED, COMPUTE THE DETERMINANT.
 CHECK = PRESENT(D)

 SELECT CASE (CHECK)

	!IF DETERMINANT IS REQUIRED: 
	 CASE (.TRUE.)

		!IF GIX IS SINGULAR, DETERMINANT = 0
		 IF(RANK<N) THEN
			 D = 0.D0
			 CONTINUE
		 ELSE
			 D = 1.D0
			 DO i = 1, N
				 D = (GIX(i,i)**2) * D
			 ENDDO
		 ENDIF
		!---

	!ELSE, CONTINUE
	 CASE (.FALSE.)

		 CONTINUE
		!***

 END SELECT
!***

!STEP 3. OBTAIN T=INV-L.
 DO i = 1, N

	!iTH DIAGONAL ENTRY IS ZERO...
	 IF (GIX(i,i).EQ.0.D0) THEN

	 GIX(i+1:N,i) = 0.D0 !SET THE OTHER ELEMENTS IN	THE CORRESPONDING COLUMN TO ZERO.
		 
	!iTH DIAGONAL ENTRY IS NOT ZERO...	 
	 ELSE

 	 GIX(i,i) = 1.D0 / GIX(i,i) !T(i,i)=1/L(i,i) 

	 W(i+1:N) = 0 !SET WORKING VECTOR TO ZERO.

		 DO j = i+1, N
	 
			 W(j:N) = W(j:N) - GIX(j:N,j-1)*GIX(j-1,i)

			!jTH DIAGONAL ENTRY IS ZERO...
			 IF (GIX(j,j).EQ.0.D0) THEN

			 GIX(j,i) = 0.D0

			!jTH DIAGONAL ENTRY IS NOT ZERO...
			 ELSE

			 GIX(j,i) = W(j) / GIX(j,j) !T(j,i)= -SUM(k=1,j-1)[L(j,k)*T(k,i)]/L(j,j),
										!WHEN i=1,N & j=i+1,N 

			 ENDIF

		 ENDDO
		 
	 ENDIF

 ENDDO	 
!***

!STEP 4. OBTAIN LOWER TRIANGULAR ELEMENTS OF T'T=GIX.
 DO j = 1, N

	 DO i = j, N

		 GIX(i,j) = DOT_PRODUCT(GIX(i:N,j), GIX(i:N,i))
	 
	 ENDDO

 ENDDO
!***

!STEP 4. COMPUTE ABOVE DIAGONAL ELEMENTS BY SYMMETRY
 DO i = 1, N

	 GIX(i,i+1:N) = GIX(i+1:N,i)
	 
 ENDDO
!***

 END SUBROUTINE GINV
!***
!
!*****************************************************************************************
!SUBROUTINE CHOLESKY--------------------------------------------------- BY SML. 16/06/2009
!*****************************************************************************************

!PERFORM CHOLESKY DECOMPOSITION OF A X(N,N) POSITIVE DEFINITE SYMMETRIC MATRIX. OPTIONALLY
!LINEAR FORM, ROW ORDERED STORAGE FOR LOWER TRIANGULAR MATRIX L IS PERFORMED. THIS STORAGE
!STRUCTURE IS EQUIVALENT TO A COLUMN ORDERED STORAGE IF L WHERE UPPER TRIANGULAR.
!***

!***VERSION 1.0***
!ADAPTED FROM SUBROUTINE GINV v1.0.

!ALGORITHM DESCRIPTION********************************************************************
!FOR A POSITIVE DEFINITE SYMMETRIC MATRIX X, X=LL' HOLDS, WITH L A LOWER TRIANGULAR MATRIX.
!ITS ENTRIES CAN BE COMPUTED USING CHOLESKY ALGORITHM.
!*****************************************************************************************

 SUBROUTINE CHOLESKY (X, N, TOL, RANK, D)

!INTENT(INOUT):  X - INPUT: TARGET MATRIX. LOWER TRIANGULAR MATRIX L.
!INTENT(IN):     N - MATRIX DIMENSION.
!INTENT(IN):   TOL - ROUNDING ERROR TOLERANCE.
!INTENT(OUT): RANK - MATRIX RANK.
!INTENT(OUT):    D - OPTIONALLY, STORE L IN A LINEAR FORM, ROW ORDERED STRUCTURE.

!********************
!VARIABLE DECLARATION
!********************
 IMPLICIT NONE !ALL VARIABLES MUST BE DECLARED.

!DUMMY ARGUMENTS
 INTEGER(K4B),INTENT(IN):: N
 INTEGER(K4B),INTENT(OUT):: RANK
 REAL(DP),INTENT(IN):: TOL
 REAL(DP),INTENT(INOUT):: X(N,N)
 REAL(DP),INTENT(OUT), OPTIONAL:: D(N*(N+1)/2)

!LOCAL VARIABLES
 LOGICAL:: CHECK !.TRUE. IF LINEAR FORM ROW ORDERED STRUCTURE IS REQUIRED.
 INTEGER(K4B):: i, j, k !DUMMY VARIABLES.
!***

!*********
!PROCEDURE
!*********

 RANK = N

!PERFORM CHOLESKY DECOMPOSITION. OBTAIN LOWER TRIANGULAR MATRIX L, SUCH THAT LL'=X.
!OVERWRITES BELOW DIAGONAL ELEMENTS OF X.
 DO i = 1, N

	 DO j = 1, i-1 

		 X(i:N,i) = X(i:N,i) - X(i,j)*X(i:N,j)
													   
	 ENDDO

	!iTH DIAGONAL ENTRY IS ZERO...
	 IF (X(i,i).LT.TOL) THEN

		 RANK = RANK-1 !FOUND A LINEAR DEPENDENCY.

	!iTH DIAGONAL ENTRY IS NOT ZERO...	 
	 ELSE

 		 X(i,i) = DSQRT(X(i,i)) !L(1,1)=SQRT[X(1,1)], WHEN i=1.
							    !L(i,i)=SQRT{X(i,i)-SUM(j=1,i-1)[L(i,j)^2]}, WHEN i=2,N.

		 X(i+1:N,i) = X(i+1:N,i)/X(i,i) !L(k,1)=X(k,1)/L(1,1), WHEN i=1 & k=2,N
		 								!L(k,i)={X(k,i)-SUM(j=1,i-1)[L(i,j)*L(k,j)]}/L(i,i),
  										!WHEN i=2,N & k=i+1,N
	 ENDIF

 ENDDO	 
!***		 


!ESTABLISH IF LINEAR FORM, ROW ORDERED STRUCTURE IS REQUIRED 
 CHECK = PRESENT(D)


!STORE RESULTS
 SELECT CASE (CHECK)

	!IF LINEAR FORM, ROW ORDERED STRUCTURE IS REQUIRED: 
	 CASE (.TRUE.)

		 k = 0
		 DO j = 1, N
			 DO i = 1, j
				k = k + 1
				D(k) = X(j,i)
			 ENDDO
		 ENDDO
		!***

	!ELSE, MATRIX FORM STORAGE IS 
	 CASE (.FALSE.)

		 DO i = 1, N
			 X(i,i+1:N) = 0.D0
		 ENDDO
		!***

 END SELECT
!***

 END SUBROUTINE CHOLESKY
!***

 END MODULE MALG