!----------------
 PROGRAM RDBLK_GS
!----------------
!
!DATE: 07/05/2015
!***
!
!THIS PROGRAM WAS SPECIFICALLY  WRITTEN TO SET-UP AND SOLVE SINGLE-TRAIT MATERNAL ANIMAL MIXED MODEL
!EQUATIONS WHEN DAM-OFFSPRING RESIDUALS ARE CORRELATED AND, THUS, THE ERROR  COVARIANCE MATRIX R HAS
!A BLOCK DIAGONAL STRUCTURE. FURTHER, IT CREATES A MARKOV CHAIN VIA ITERATIVE MONTE CARLO SIMULATION
!USING A SINGLE-SITE SYSTEMATIC GIBBS SAMPLER ALGORITHM. ULTIMATELY, THE MAIN INTEREST IS TO ESTIMA-
!TE THE (CO)VARIANCE COMPONENTS UNDER A BAYESIAN APPROACH.
!
!***VERSION 1.1***
!UPDATES VERSION 1.0 BY NEW PRECISION PARAMETERIZATION AND NEW PROCEDURES DEPENDENCES.
!
!-------------------
!ADDITIONAL FEATURES
!-------------------
! ADMITS BOTH, CATEGORICAL FACTORS & (NESTED) COVARIATES FOR THE FIXED EFFECTS.
! SUPPORTS MULTIPLE INCIDENCE MATRICES, INDENTICALLY AND INDEPENDENTLY DISTRIBUTED (IID) RANDOM EF-
!  FECTS AND RANDOM ANIMAL EFFECTS (INBRED POPULATIONS).
! SUPPORTS POSITIVE SEMI DEFINITE SYSTEMS, ALTHOUGH A REPARAMETERIZATION BLOCK IS AVAILABLE AND RE-
!  COMMENDED TO AVOID NUMERICAL INACCURACIES. ITS ALSO POSSIBLE TO DEFINE AN EFFECT AS A CONTRAST U-
!  SING NEGATIVE VALUES IN THE DATA FILE.
! THEORETICAL BACKGROUND ON BAYESIAN STATISTICS AND GIBBS SAMPLING ALGORITHM BASED ON SORENSEN AND
!  GIANOLA'S BOOK (2002, CH.13)
!
!---------------
!TECHNICAL NOTES
!---------------
! SPARSE STORAGE OF THE EQUATION'S SYSTEM.
! USES "CLPS" MODULE TO COLLAPSE A SYSTEM WITH ZERO ROWS AND COLUMNS. 
! USES AN [OPTIONALLY] ADAPTIVE, THOUGH NO GROWER, GRIDDY-GIBBS ALGORITHM TO SAMPLE FROM THE UNKNOWN
!  FULL CONDITIONAL DENSITY OF THE CORRELATION PARAMETER. 
!
!------------------------------------
!BUGS - ATT. BEWARE SOFTWARE PITFALLS
!------------------------------------
! COMPUTER RUN OUT OF MEMORY AND PROGRAM STOPPED AROUND ITERATION 488,000 IN A 3-BLOCK 11-INDIVIDUAL
!  EXAMPLE, WITH 9 GRID-VALUES. (DEBUGGED 17-02-2010).
!***
!
!AUTHOR: SML * BASED ON THE "COMPUTATIONAL TECHNIQUES IN ANIMAL BREEDING" CLASS NOTES FROM PROFESSOR
!IGNACY MISTZAL, UNIVERSITY OF GEORGIA, ATHENS, USA. CODE ADAPTED FROM PROGRAM 'BLUP v.3.2.1.1'.
!***
!
!ATT. THIS SYMBOL MEANS "ATTENTION: CHANGE ACCORDING TO THE SPECIFIC MODEL TO BE FITTED".
!***
!
!ATT. COMPILE WITH "PARAM", "ITPACK", "MALG", "SIM", "CLPS" & "SAMPLER v2.0" MODULES.
!***
!
!************
!USED MODULES
!************
 USE PARAM
 USE ITPACK
 USE MALG
 USE SIM
 USE CLPS
 USE SAMPLER
!
!**********************************
!SYMBOL DEFINITION (ALPHABETICALLY)
!**********************************
!  ADDJ: SCALAR VARIABLE USED TO ADDRESS CONTRIBUTIONS FOR iid RANDOM EFFECTS. 
! ADDJJ: COEFFICIENT MATRIX ROW NUMBER, USED TO ADDRESS CONTRIBUTIONS FOR RANDOM ANIMAL EFFECTS. 
! ADDKK: COEFFICIENT MATRIX COLUMN NUMBER, USED TO ADDRESS CONTRIBUTIONS FOR RANDOM ANIMAL EFFECTS.
!ADRESS: MATRIX FUNCTION OF (FACTOR, LEVEL, WITHIN-BLOCK INDIVIDUAL) USED TO ADDRESS CONTRIBUTIONS
!        FOR EACH DATA POINT. 
!  AMAT: SPARSE_MATRIX DERIVED TYPE VARIABLE, USED TO STORE A-INVERSE MATRIX.
!	ANZ: ACTUAL NUMBER OF NONZEROS.
!  AVEC: REAL, DOUBLE PRECISION VECTOR(NZ). STORES THE VALUE OF THE NONZERO ENTRIES.
!  BETA: VECTOR(NEQ) CONTAINING THE SOLUTIONS OF THE MIXED MODEL EQUATIONS.
!  CLSS: CATEGORICAL FACTOR CODE IDENTIFIER.
!  CORR: VARIABLE USED TO ASCERTAIN CORRELATIONS BETWEEN RANDOM EFFECTS.
!  COVR: COVARIATE CODE IDENTIFIER.
!  CVRW: COVARIATE WEIGHT.
!  DCOL: INTEGER WORK VECTOR(NEQ) USED WITHIN COLLAPSE SUBROUTINE.
!  DET1: Rk-MATRIX DETERMINANT.
!  DETR: R-MATRIX DETERMINANT.
!  EFFT: VECTOR WITH THE CATEGORICAL (CLSS) OR COVARIABLE (COVR) CODES OF EACH EFFECT.
!	ESP: INTEGER WORK VARIABLE USED WITHIN SIM'S MODULE.
!	 FK: NUMBER OF RECORDED INDIVIDUALS WITHIN EACH BLOCK
!  FLAG: ERROR FLAG WITHIN SIM MODULE.
!  FIXX: FIXED EFFECT CODE INDENTIFIER.
!     G: RANDOM EFFECTS COVARIANCE MATRICES. TRIDIMENSIONAL. FIRST INDEX POINTS TO THE EFFECT.
!	 IA: VECTOR(NEQ+1). POINTS TO THE FIRST ENTRY OF EACH ROW OF THE COEFFICIENT MATRIX.
!    IP: VECTOR(NEQ) USED TO RETURN THE INVERSE OF THE PERMUTATION RETURNED IN P.
!  IERR: ERROR FLAG WITHIN MODULE ITPACK'S SBSIJ SUBROUTINE.
!INDATA: INCIDENCE VECTOR OF EFFECTS ON A DATA LINE. 
!   ISP: VECTOR(NSP) USED FOR WORKING STORAGE WITHIN SIM'S MODULE.
! IWORK: WORK VECTOR(NZ) USED TO TEMPORARILY STORE LINKED LISTS.
!	 JA: VECTOR(NZ). STORES COLUMN NUMBER OF EACH NON ZERO ELEMENT OF THE COEFFICIENT MATRIX.
!  MAXX: MAXIMUM VALUE POSSIBLE FOR THE DIMENSION OF A RANDOM EFFECT COVARIANCE MATRIX.
!  MISS: MISSING TRAIT CODE IDENTIFIER.
!  MODE: SPECIFIES THE WAY IN WHICH THE ENTRY IS TO BE TREATED, WHEN (I,J) ENTRY HAS ALREADY BEEN SET.
!     N: NUMBER OF ANIMALS IN THE PEDIGREE FILE.
!  NBKS: NUMBER OF BLOCKS IN R-MATRIX.
!  NEFF: NUMBER OF FACTORS OR EFFECTS INCLUDING COVARIATES.
!   NEQ: NUMBER OF EQUATIONS.
!  NEST: NESTED COVARIABLE CODE IDENTIFIER.
!  NLEV: NUMBER OF LEVELS OF EACH FACTOR.
! NOUTT: OUTPUT FILE UNIT NUMBER WITHIN MODULE ITPACK'S SBSIJ SUBROUTINE.
!  NRDF: NUMBER OF RECORDS IN THE DATA FILE.
!	NSP: DECLARED DIMENSION OF THE SIM'S MODULE WORK VECTOR ISP. MUST BE AT LEAST 3*NEQ+4*NZ.
!	 NZ: NUMBER OF NON ZEROS IN THE COEFFICIENT MATRIX.
!  NZRW: NON ZERO ROW INDICATOR.
!	  P: PERMUTATION VECTOR(NEQ) USED TO RETURN THE OPTIMAL ORDERING OF COEFFICIENT MATRIX IN SIM'S MODULE.
! PATH1: PATH SPECIFICATION WITHIN ODRV SUBROUTINE.
! PATH2: PATH SPECIFICATION WITHIN SDRV SUBROUTINE.
!  pPAR: CORRELATION AMONG ERRORS PARAMETER. pPAR0: INITIAL VALUE.
!     R: Rk(:,:) ALLOCATABLE COVARIANCE MATRIX FOR EACH BLOCK OF ERRORS.
!  RANK: Rk RANK.
!   RES: WITHIN-BLOCK RESIDUALS MATRIX.
!  RINV: INV-Rk.
!  RND1: INDENTICALLY AND INDEPENDENTLY DISTRIBUTED (iid) RANDOM EFFECT CODE IDENTIFIER.
!  RND2: RANDOM ANIMAL EFFECT CODE IDENTIFIER.
!   RSP: VECTOR(NSP) USED FOR WORKING STORAGE WITHIN SIM'S MODULE.
!  S2ER: WHITE-NOISE ERROR VARIANCE PARAMETER. S2E0: INITIAL VALUE.
!	SSR: RESIDUAL QUADRATIC FORM: SUM(k)[RES(K)'*RINV(K)*RES(K)] 
! START: MARKS POSITION OF (FACTOR, LEVEL) COMBINATION FOR EACH DATA POINT ON X'(INV-R)X MATRIX.
!SWITCH: OUTPUT FILE SWITCH WITHIN MODULE ITPACK'S SBSIJ SUBROUTINE. 
!  TRCK: LOGICAL MATRIX, USED TO KEEP TRACK OF DEFINED EFFECTS FOR EACH TRAIT.
!  TYPO: VECTOR WITH THE FIXED (FIXX) OR RANDOM (RND1, RND2) CODES OF EACH EFFECT.
!  VALL: SCALAR VARIABLE WITH THE CONTRIBUTION VALUE WITHIN EACH CALL OF MODULE ITPACK. 
!    XY: RHS VECTOR.
!     Y: (ALLOCATABLE) STORE DATA FOR EACH BLOCK.
!  ZROW: BINARY VECTOR(NEQ) USED TO POINT ROWS TO BE DELETED.
!******

!********************
!VARIABLE DECLARATION
!********************
 IMPLICIT NONE !ALL VARIABLES MUST BE DECLARED.
!
!------------------
!DERIVED DATA TYPES
!------------------
!"SPARSE_MATRIX" IS USED FOR MATRICES THAT WILL BE STORED IN IJA-FORMAT.
 TYPE SPARSE_MATRIX
	 INTEGER(K4B):: N1, NZ1		!N1: N OF ROWS; NZ1: N OF NONZEROS.
     INTEGER(K4B), POINTER::   IA1(:)
     INTEGER(K4B), POINTER::   JA1(:)
	 INTEGER(K4B), POINTER::  IWK1(:)
	 REAL(DP):: VALL1				
	 REAL(DP), POINTER:: A1(:)
 END TYPE
!
!"BLOCK" IS USED FOR VARIABLES ASSOCIATED TO A PARTICULAR R_BLOCK.
 TYPE BLOCK
	 INTEGER(K4B):: FK !FK: BLOCK SIZE.
	 INTEGER(K4B), POINTER:: ADRESS(:,:)
	 REAL(DP), POINTER:: Y(:), RES(:)
     REAL(DP), POINTER:: INDATA(:,:), CVRW(:,:), R(:,:), RINV(:,:)
 END TYPE
!---
!
!---------------------------
!TYPE DECLARATION STATEMENTS
!---------------------------
 INTEGER(K4B):: NEQ, START
 INTEGER(K4B), PARAMETER:: MAXX=2      !ATT. DEFINE MAXIMUM VALUE POSSIBLE FOR CORR(EFFECT).
 INTEGER(K4B), PARAMETER:: FIXX=0
 INTEGER(K4B), PARAMETER:: RND1=1
 INTEGER(K4B), PARAMETER:: RND2=2
 INTEGER(K4B), PARAMETER:: CLSS=0
 INTEGER(K4B), PARAMETER:: COVR=1
 INTEGER(K4B), PARAMETER:: MISS=0
 INTEGER(K4B), PARAMETER:: MODE=5
!---
!
!-----------------------
!FIXED EFFECTS VARIABLES
!-----------------------
 INTEGER(K4B):: NBKS
 REAL(DP), ALLOCATABLE:: XY(:)
 TYPE(BLOCK), ALLOCATABLE:: BK(:)
!
!------------------------------------
!ITPACK, SIM & CLPS MODULES VARIABLES
!------------------------------------
 INTEGER(K4B), ALLOCATABLE:: IA(:), JA(:), IWORK(:), P(:), IP(:), ISP(:), ZROW(:), DCOL(:)
 INTEGER(K4B):: SWITCH, NOUTT, IERR, NSP, PATH1, PATH2, FLAG, ESP
 REAL(DP), ALLOCATABLE:: AVEC(:), BETA(:), RSP(:)
 REAL(DP):: VALL				
!---
!
!-----------------------
!GIBBS SAMPLER VARIABLES
!-----------------------
!LOCATION PARAMETERS VARIABLES
 REAL(DP), ALLOCATABLE:: TITA(:)
!
!RANDOM EFFECTS COVARIANCE VARIABLES
 TYPE(SPARSE_MATRIX):: AMAT
!---
!
!-------------------
!OPERATIVE VARIABLES
!-------------------
 INTEGER(K4B):: CUM1, ANZ, CONT, COUNT2, COUNT3, ERROR !COUNTING VARIABLES.
 INTEGER(K4B):: BLK, i, ii, j, k, l, ROUND !DUMMY VARIABLES.
 INTEGER(K4B):: RANK, ADDJ, NZRW !WORKING VARIABLES.
!---
!
!*********
!PROCEDURE
!*********
 PRINT*, '---------------------'
 PRINT*, 'PROGRAM RDBLK_GS v1.1'
 PRINT*, '---------------------'
 PRINT*, ''

!DECODE INPUT PARAMETER FILE
 CALL DECODE_PARAM_FILE
 PRINT*, 'PARAMETER FILE SUCCESFULLY READ. START EXECUTION...'
 PRINT*, ''
!---
!
!*********************************************************
!1. INITIALIZATION OF MATRICES, VECTORS & SCALAR VARIABLES
!*********************************************************
 PRINT*, 'STEP 1. INITIALIZATION.'
 PRINT*, ''
!
!----------------------------------
!SET ITPACK AND SIM MODULES OPTIONS
!----------------------------------
!ATT. SET MODULE ITPACK'S SBSIJ SUBROUTINE OUTPUT FILE OPTIONS.
 SWITCH= 0 !ATT. SET AN INTEGER GREAT OR EQUAL "0" FOR PRINTING THE OUTPUT FILE.
  NOUTT= 6 !ATT. SET THE UNIT NUMBER OF THE OUTPUT FILE. SET '6' TO PRINT ON SCREEN.
!
!ATT. SET MODULE SIM'S ODRV SUBROUTINE PATH OPTION.
 PATH1= 2 !ATT. SET "2" TO FIND OPTIMAL ORDERING AND REORDER SYMMETRICALLY STORED MATRIX. OTHER
		  !OPTIONS ARE AVAILABLE (SEE LINE "SIM00990" IN SIM'S MODULE).
 PATH2= 1 !ATT. SET "1" TO PERFORM THE THREE STAGES OF SDRV. OTHER OPTIONS ARE AVAILABLE (SEE
		  !LINE "SIM08480" IN SIM'S MODULE). 	
!---
!
!--------------------------------------
!ALLOCATE BOUNDS & INITIALIZE VARIABLES
!--------------------------------------
 NEQ = SUM(NLEV)

 ALLOCATE (XY(NEQ), IA(NEQ+1), JA(NZ), IWORK(NZ), AVEC(NZ), ZROW(NEQ), DCOL(NEQ), BETA(NEQ))

 CALL SBINI(NEQ,NZ,IA,JA,AVEC,IWORK)

 XY = 0; NBKS = 0; BETA = 0
!---
!
!-------------------------------------------------------------
!INITILIZATION OF DERIVED-DATA TYPE "SPARSE_MATRIX" COMPONENTS
!-------------------------------------------------------------
!DISASSOCIATE POINTERS VARIABLES 
 NULLIFY(AMAT%IA1, AMAT%JA1, AMAT%IWK1, AMAT%A1)

!SET AMAT MATRIX COMPONENTS AND ALLOCATE SPARSE-STORAGE VARIABLES 
 AMAT%N1    = N
! AMAT%NZ1   = 50000 !ATT. SET THE ESTIMATED NUMBER OF NON ZERO ENTRIES FOR GMAT MATRICES
 AMAT%NZ1   = INT(NZ*0.1D0)
 AMAT%VALL1 = 0.D0

 ALLOCATE(AMAT%IA1(AMAT%N1+1), AMAT%JA1(AMAT%NZ1), AMAT%IWK1(AMAT%NZ1), AMAT%A1(AMAT%NZ1), STAT=ERROR)

	 IF(ERROR.NE.0) STOP "AMAT ALLOCATION ERROR" !CHECK FOR ALLOCATION ERROR.

 CALL SBINI(AMAT%N1,AMAT%NZ1,AMAT%IA1,AMAT%JA1,AMAT%A1,AMAT%IWK1)
!---
!
!-------------------------------------------------------------
!REORDER INPUT DATA FILE: CALL INTERNAL SUBROUTINE "FILE_PREP"
!-------------------------------------------------------------
 CALL FILE_PREP
!---
!
!-----------------------------------------------------
!INITILIZATION OF DERIVED-DATA TYPE "BLOCK" COMPONENTS
!-----------------------------------------------------
!ALLOCATE POINTERS VARIABLES 
 DO i = 1, NBKS

	 NULLIFY(BK(i)%ADRESS, BK(i)%Y, BK(i)%RES, BK(i)%INDATA, BK(i)%CVRW, BK(i)%R, BK(i)%RINV)

	 ALLOCATE(BK(i)%ADRESS(NEFF,BK(i)%FK), BK(i)%Y(BK(i)%FK), BK(i)%RES(BK(i)%FK),&
			  BK(i)%INDATA(NEFF,BK(i)%FK), BK(i)%CVRW(NEFF,BK(i)%FK), BK(i)%R(BK(i)%FK,BK(i)%FK),&
			  BK(i)%RINV(BK(i)%FK,BK(i)%FK), STAT=ERROR)

	 IF(ERROR.NE.0) STOP "BLOCK ALLOCATION ERROR" !CHECK FOR ALLOCATION ERROR.

	 BK(i)%ADRESS=0; BK(i)%Y=0; BK(i)%RES=0; BK(i)%INDATA=0; BK(i)%CVRW=0; BK(i)%R=0; BK(i)%RINV=0

 ENDDO
!---
!
!--------------------
!OPEN OPERATIVE FILES
!--------------------
 OPEN (21, FILE='DAT1.sal',STATUS='OLD')

!ATT. NAMES OF OUTPUT FILES
 OPEN (25, FILE='CVC-CHAIN.txt',STATUS='REPLACE')   !ATT. NAME OF COVARIANCE COMPONENTS OUTPUT FILE.
!---
!
 PRINT*, ''
 PRINT*, 'NUMBER OF EQUATIONS:', NEQ
 PRINT*, '   NUMBER OF BLOCKS:', NBKS
 PRINT*, ''
!
!******************************
!2. FIXED EFFECTS CONTRIBUTIONS
!******************************
 PRINT*, '***'
 PRINT*, 'STEP 2. FIXED EFFECTS CONTRIBUTIONS.'
 PRINT*, ''
 COUNT3 = 1
!
!------------------
!CYCLE AMONG BLOCKS
!------------------
 DO BLK = 1, NBKS

	!-------------------------
	!DATA READING CHECK BLOCK
	!-------------------------
     IF(BLK == (COUNT3*100)) THEN
		 WRITE(*,'(i9,2x,a13)') BLK, 'FAMILIES READ'
		 COUNT3 = COUNT3 + 1
	 ENDIF
	!---
	!
	!-------------------------
	!READ AND STORE DATA BLOCK
	!-------------------------
	 DO i = 1, BK(BLK)%FK 
	  	 READ(21, *) BK(BLK)%INDATA(:,i), BK(BLK)%Y(i)
	 ENDDO
	!---
	!
	!--------------
	!ADRESS SETTING
	!--------------
	 CALL SET_ADDRESS 
	!---
	!
	!---------------------
	!SET-UP Rk-MATRIX*S2E0 
	!---------------------
	!ATT. CHANGE STATEMENTS ACCORDING TO THE R-STRUCUTURE  
	 DO i = 1, BK(BLK)%FK

		 BK(BLK)%R(i,i) = S2E0 !ADDS VARIANCE TO THE DIAGONAL

		 DO j = i+1, BK(BLK)%FK
	 	 
			!IF j IS DAM OF i:
			 IF (BK(BLK)%INDATA(NEFF,j).EQ.BK(BLK)%INDATA(NEFF-1,i)) THEN
			 BK(BLK)%R(i,j) = pPAR0*S2E0     !ADDS Ro PARAMETER TO THE CORRESPONDING ENTRY
			 BK(BLK)%R(j,i) = BK(BLK)%R(i,j) !TRANSPOSE RESULT
			 ENDIF

		 ENDDO

	 ENDDO
	!---
	!
   	!---------------
	!INVERT R-MATRIX
	!---------------
	 BK(BLK)%RINV = BK(BLK)%R

	 IF(BLK.NE.NBKS) THEN
	 
		CALL GINV(BK(BLK)%RINV,BK(BLK)%FK,BK(BLK)%FK,1.D-12,RANK)

			!SECURITY DEVICE: CHECK OUT IF SCALE MATRIX IS POSITIVE DEFINITE.
			 IF(RANK .LT. BK(BLK)%FK) THEN
				 PRINT*, ''
				 PRINT*, '-----------'
				 WRITE(*,'(a30,i3,a27)') 'FATAL ERROR: RINV FOR FAMILY ', BLK,&
										 ' IS NOT POSITIVE DEFINITE.' 
				 PRINT*, '-----------'
				 PRINT*, ''
				  STOP
			 ENDIF

	 ELSE

		 DO i = 1, BK(BLK)%FK	
			 BK(BLK)%RINV(i,i) = 1.D0 / BK(BLK)%R(i,i)
		 ENDDO

	 ENDIF	
	!---
	!
   	!--------------------------------------
	!COEFFICIENT MATRIX BLOCK CONTRIBUTIONS
	!--------------------------------------
	 DO i = 1, NEFF
		 DO k = 1, BK(BLK)%FK

			 DO j = 1, NEFF

				!EXCEPT FOR THE LAST BLOCK
				 IF(BLK.NE.NBKS) THEN

					 DO l = 1, BK(BLK)%FK

						!ONLY COMPUTE UPPER DIAGONAL AND NONZERO CONTRIBUTIONS
						 IF(BK(BLK)%ADRESS(i,k).LE.BK(BLK)%ADRESS(j,l) .AND.&
							BK(BLK)%ADRESS(i,k).NE.0 .AND. BK(BLK)%ADRESS(j,l).NE.0) THEN
					 
							 VALL = BK(BLK)%CVRW(i,k)*BK(BLK)%CVRW(j,l)*BK(BLK)%RINV(k,l) 

							 CALL SBSIJ(NEQ,IA,JA,AVEC,IWORK,BK(BLK)%ADRESS(i,k),&
										BK(BLK)%ADRESS(j,l),VALL,MODE,SWITCH,NOUTT,IERR)
						 
							!AN ERROR HAS OCCURRED. STOP PROCESS.
							 IF (IERR==701 .OR. IERR==702) STOP 
						 
						 ENDIF

					 ENDDO

				!FOR THE LAST BLOCK, WITH IID ERROR STRUCTURE
				 ELSE

					!ONLY COMPUTE UPPER DIAGONAL AND NONZERO CONTRIBUTIONS
					 IF(BK(BLK)%ADRESS(i,k).LE.BK(BLK)%ADRESS(j,k) .AND. BK(BLK)%ADRESS(i,k).NE.0&
					    .AND. BK(BLK)%ADRESS(j,k).NE.0) THEN
					 
						 VALL = BK(BLK)%CVRW(i,k)*BK(BLK)%CVRW(j,k)*BK(BLK)%RINV(k,k) 
						 CALL SBSIJ(NEQ,IA,JA,AVEC,IWORK,BK(BLK)%ADRESS(i,k),BK(BLK)%ADRESS(j,k),&
									VALL,MODE,SWITCH,NOUTT,IERR)
			 
						!AN ERROR HAS OCCURRED. STOP PROCESS.
						 IF (IERR==701 .OR. IERR==702) STOP 

					 ENDIF
						
				 ENDIF
				!---	 		 	

			 ENDDO

			!-----------------
			!RHS CONTRIBUTIONS
			!-----------------
			!EXCEPT FOR THE LAST BLOCK
			 IF(BLK.NE.NBKS) THEN

				 IF (BK(BLK)%ADRESS(i,k)==0) CYCLE

				 DO l = 1, BK(BLK)%FK
					 XY(BK(BLK)%ADRESS(i,k)) = XY(BK(BLK)%ADRESS(i,k)) + BK(BLK)%Y(l)*&
											   BK(BLK)%CVRW(i,k)*BK(BLK)%RINV(k,l)
				 ENDDO
				!---

			!FOR THE LAST BLOCK, WITH IID ERROR STRUCTURE
			 ELSE

				 IF (BK(BLK)%ADRESS(i,k)==0) CYCLE

				 XY(BK(BLK)%ADRESS(i,k)) = XY(BK(BLK)%ADRESS(i,k)) + BK(BLK)%Y(k)*&
										   BK(BLK)%CVRW(i,k)*BK(BLK)%RINV(k,k)
			 ENDIF
			!---

		 ENDDO 
	 ENDDO 
	!---

 ENDDO
!---
!
!*******************************
!3. RANDOM EFFECTS CONTRIBUTIONS
!*******************************
 PRINT*, '***'
 PRINT*, 'STEP 3. RANDOM EFFECTS CONTRIBUTIONS.'
 PRINT*, ''
!
!-------------------
!CYCLE AMONG EFFECTS
!-------------------
 DO ii = 1, NEFF

	!-----------------------------------------
	!INVERT RANDOM EFFECTS COVARIANCE MATRICES
	!-----------------------------------------
	 INVG = G0
	 IF (CORR(ii).NE.0) THEN

		 CALL GINV(INVG(ii,:,:),CORR(ii),MAXX,1.D-12,RANK)

		!SECURITY DEVICE: CHECK OUT IF G0(ii,:,:) IS POSITIVE DEFINITE.
		 IF(RANK .LT. CORR(ii)) THEN
			 PRINT*, ''
			 PRINT*, '-----------'
			 WRITE(*,'(a24,i2,a27)') ' FATAL ERROR: MATRIX G0(', ii, ') IS NOT POSITIVE DEFINITE.' 
			 PRINT*, '-----------'
			 PRINT*, ''
			  STOP
		 ENDIF

	 ELSE
		 CYCLE
	 ENDIF
	!---
	!
	!-----------------------------
	!SELECT THE RANDOM EFFECT TYPE
	!-----------------------------
	 SELECT CASE(TYPO(ii))
	
 		!IF IT IS A FIXED EFFECT:
		 CASE (FIXX)

			 CONTINUE !SKIP EFFECT

		!IF IT IS A RANDOM iid EFFECT:
		 CASE (RND1)
			
			 START = SUM(NLEV(1:ii-1)) + 1

			 DO j = 1, NLEV(ii)

				!ADDRESS SETTING.
				 ADDJ =  START + (j - 1)

				!ENTRIES CALCULATION.
			     CALL SBSIJ(NEQ,IA,JA,AVEC,IWORK,ADDJ,ADDJ,INVG(ii,1,1),MODE,SWITCH,NOUTT,IERR)

				!AN ERROR HAS OCCURRED. STOP PROCESS.
				 IF (IERR==701 .OR. IERR==702) STOP 

			 ENDDO
			!---

		!IF IT IS A RANDOM ANIMAL EFFECT:
		 CASE (RND2)

			!COMPUTE SINGLE TRAIT INVERSE COVARIANCE MATRIX & ADD CONTRIBUTIONS TO WRW.
			 CALL INVA(ii)

		!ELSE:
		 CASE DEFAULT

			 PRINT*, 'UNIMPLEMENTED RANDOM EFFECT', TYPO(ii)
			 STOP
		
	 END SELECT
	!---

 ENDDO
!---
!
!*****************************
!4. BUILD FINAL DATA STRUCTURE
!*****************************
 PRINT*, '***'
 PRINT*, 'STEP 4. BUILDING FINAL DATA STRUCTURE.'
 PRINT*, ''
!
!-----------------------
!SEARCHING FOR ZERO ROWS
!-----------------------
 NLEV1 = NLEV 
 ZROW  = 0
 DCOL  = 0
 CUM1  = 0
 NZRW  = 0
 CONT  = 0

!SEARCH AND TAG ZERO ROWS
 DO i = 1, NEFF
	 DO j = 1, NLEV(i)
		 CONT = CONT + 1
		 IF (IA(CONT).LE.0) THEN
			 NZRW = 1
			 CUM1 = CUM1 + 1 
			 ZROW(CONT) = 1
			 NLEV1(i) = NLEV1(i)-1 !UPDATING NLEV VECTOR.
		 ENDIF
		 DCOL(CONT) = CUM1
	 ENDDO	
 ENDDO
!---
!
!-----------------------------------------
!BUILD SPARSE-STORAGE FINAL DATA STRUCTURE
!-----------------------------------------
 CALL SBEND(NEQ,NZ,IA,JA,AVEC,IWORK)
 CALL SBEND(AMAT%N1,AMAT%NZ1,AMAT%IA1,AMAT%JA1,AMAT%A1,AMAT%IWK1) 
!---
!
!-------------------
!COLLAPSE THE SYSTEM
!-------------------
 IF (NZRW.NE.0) THEN
	 PRINT*, ''
	 PRINT*, 'ZERO ROWS FOUND. COLLAPSING THE SYSTEM...'
	 CALL COLLAPSE(NEQ, IA, JA, XY, ZROW, DCOL(1:NEQ))
	 PRINT*, 'REMAINING EQUATIONS:', NEQ
 ENDIF
!
!TEST FOR UPDATED NLEV VECTOR
 IF(NEQ.NE.SUM(NLEV1)) STOP "CHECK NLEV UPDATING" 
!---
!
!-----------------------------------------
!SET THE ACTUAL NUMBER OF NON ZERO ENTRIES
!-----------------------------------------
 PRINT*, ''
 PRINT*, 'SETTING THE ACTUAL NUMBER OF NONZERO ENTRIES...'

 ANZ = IA(NEQ+1) - 1
 PRINT*, 'No. NONZERO ENTRIES:', ANZ
!----
!
!************************************
!5. SOLUTION OF MIXED MODEL EQUATIONS
!************************************
 PRINT*, ''
 PRINT*, '***'
 PRINT*, 'STEP 5. SOLVING MIXED MODEL EQUATIONS.'
 PRINT*, ''
!
!-----------------------------------------------
!INITIALLIZE AND ALLOCATE SOLVER'S ARRAYS BOUNDS
!-----------------------------------------------
 NSP = 3*NEQ + 6*ANZ
 ALLOCATE (P(NEQ), IP(NEQ), ISP(NSP), RSP(NSP))
!---
!
!----------------
!SOLVE THE SYSTEM
!----------------
!SEARCH FOR OPTIMAL ORDERING.
 CALL ODRV(NEQ,IA(1:NEQ+1),JA(1:ANZ),AVEC(1:ANZ),P,IP,NSP,ISP,PATH1,FLAG)

	!AN ERROR HAS OCCURRED. STOP PROCESS.
	 IF (FLAG.NE.0) THEN
		 PRINT*, 'ERROR IN ODRV'
		 STOP
	 ENDIF
!
!SOLVE EQUATION'S SYSTEM.
 CALL SDRV(NEQ,P,IP,IA(1:NEQ+1),JA(1:ANZ),AVEC(1:ANZ),XY(1:NEQ),BETA(1:NEQ),NSP,ISP,RSP,ESP,PATH2,FLAG)

	!AN ERROR HAS OCCURRED. STOP PROCESS.
	 IF (FLAG.NE.0) THEN
		 PRINT*, 'ERROR IN SDRV'
		 PRINT*, 'FLAG:', FLAG
	     STOP
	 ENDIF
!
!*****************
!6. GIBBS SAMPLING
!*****************
 PRINT*, '***'
 PRINT*, 'STEP 6. SAMPLING.'
 PRINT*, ''
!
!----------------------------------
!ALLOCATE LOCATION PARAMETER VECTOR 
!----------------------------------
 ALLOCATE(TITA(NEQ)); TITA = BETA
!---
!
!------------------------------------------------------------
!INITIALIZE COVARIANCE STRUCTURE PARAMETERS BEFORE ITERATIONS
!------------------------------------------------------------
    G = G0
 S2ER = S2E0
 pPAR = pPAR0
!---
!
!--------------------
!GIBBS SAMPLER CYCLES
!--------------------
 COUNT2 = 1

 DO ROUND = 1, NROUND

	!------------------------------
	!PRINT THE NUMBER OF ITERATIONS 
	!------------------------------
     IF(ROUND.eq.(COUNT2*1)) THEN
         WRITE(*,'(i9,2x,a12)') ROUND, 'TH ITERATION'
         COUNT2 = COUNT2 + 1
	 ENDIF
	!---
	!
	!--------------------------------
	!COMPUTE UPDATED INVERSE G-MATRIX
	!--------------------------------
	 INVG = G !SET INVG WITH UPDATED G MATRIX

	 DO ii = 1, NEFF

		!INVERT UPDATED RANDOM EFFECTS COVARIANCE MATRICES.
		 IF (CORR(ii).NE.0) THEN

			 CALL GINV(INVG(ii,:,:),CORR(ii),MAXX,1.D-12,RANK)

			!SECURITY DEVICE: CHECK OUT IF G(ii,:,:) IS POSITIVE DEFINITE.
			 IF(RANK .LT. CORR(ii)) THEN
				 PRINT*, ''
				 PRINT*, '-----------'
				 WRITE(*,'(a23,i1,a27)') 'FATAL ERROR: MATRIX G(', ii, ') IS NOT POSITIVE DEFINITE.' 
				 PRINT*, '-----------'
				 PRINT*, ''
				  STOP
			 ENDIF

		 ELSE
			 CYCLE
		 ENDIF

	 ENDDO
	!---
	!
	!-----------------------------
	!CALL GIBBS SAMPLER SUBROUTINE
	!-----------------------------
	 CALL GIBBS(ROUND)
	!---

 ENDDO
!---
!
!---------------------
!CLOSE OPERATIVE FILES
!---------------------
 CLOSE (UNIT=21)
 CLOSE (UNIT=25)
!---
!
!ATT. COMMENT THIS STATEMENT TO PRINT OUT SOLUTIONS BEFORE FIRST ITERATION.
 IF(.NOT. PRINTOUT) STOP "PROGRAM TERMINATED - NO SOLUTIONS."
!***
!
!*********
!7. OUTPUT
!*********
 PRINT*, '***'
 PRINT*, 'STEP 6. WRITTING OUTPUT FILE.'
 PRINT*, ''
!
!--------------------------------------
!RETRIEVE SOLUTION'S ORIGINAL STRUCTURE
!--------------------------------------
 CALL RETRIEVE(NEQ, ZROW, BETA, 0.D0)
!---
!
!------------------------------------
!OPEN OUTPUT FILE AND PRINT HEADLINES
!------------------------------------
!OPEN OUTPUT FILE
 OPEN (11,file='SOLUTIONS.txt',STATUS='REPLACE') !ATT. NAME OF OUTPUT FILE. 
!
!PRINT HEADLINES
 WRITE(11,'(a14)') '++++++++++++++'
 WRITE(11,'(a14)') 'SOLUTIONS FILE'
 WRITE(11,'(a14)') '++++++++++++++'
 WRITE(11,*) ''

 WRITE(11,'(a6,2x,a5,2x,a14)') 'EFFECT', 'LEVEL', '     SOLUTIONS'
 WRITE(11,'(a6,2x,a5,2x,a14)') '------', '-----', '     ---------'

 START = 0

 DO i = 1, NEFF

	 IF (TYPO(i)==RND1) THEN

		 WRITE(11,*) ''
		 WRITE(11,'(a6,2x,a5,2x,a14)') 'EFFECT', 'LEVEL', '   PREDICTIONS'
		 WRITE(11,'(a6,2x,a5,2x,a14)') '------', '-----', '   -----------'

	 ELSEIF (TYPO(i)==RND2) THEN

		 WRITE(11,*) ''
		 WRITE(11,'(a6,2x,a5,2x,a14)') 'EFFECT', '   ID', 'BREEDING VALUE'
		 WRITE(11,'(a6,2x,a5,2x,a14)') '------', '-----', '--------------'

	 ENDIF
	
	 DO j = 1, NLEV(i)

		 START = SUM(NLEV(1:i-1)) + j

		!PRINT SOLUTIONS
		 WRITE(11,'(i6,2x,i5,2x,f14.4)') i, j, BETA(START)

	 ENDDO

	 WRITE(11,'(a29)') '-----------------------------'
	 WRITE(11,*) ''

 ENDDO
!---
!
!------------------
!CLOSE OUTPUT FILE
!------------------
 CLOSE (UNIT=11)
!---
!
 PRINT*, '***'
 STOP "PROGRAM TERMINATED."
!
 CONTAINS

!*****************************************************************************************
!INTERNAL SUBROUTINE FILE_PREP ---------------------------------------- BY SML. 29/01/2010
!*****************************************************************************************
!
!REORDERS INPUT DATA FILE RECORDS IN A WITHIN-BLOCK FASHION.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!READS AND STORES INPUT DATA FILE. THIS FILE MUST BE ORDERED SUCH THAT PARENTS ALWAYS PRE-
!CEDE THEIR OFFSPRING. ASSIGN TO EACH INDIVIDUAL A NUMERICAL CONSECUTIVE CODE ACCORDING TO
!THE BLOCK IT BELONGS. USES THE CODE TO PRINT-OUT THE OUTPUT FILE IN THE APPROPRIATE ORDER.
!*****************************************************************************************
!
!ATT. BEWARE! THIS SUBROUTINE MUST BE COMPILED WITHIN "RDBLK_GS v1.0" PROGRAM.
!
 SUBROUTINE FILE_PREP

!LOCAL VARIABLES 
 INTEGER(K4B):: DAMI !BINARY VARIABLE USED TO SET THE STATUS OF A FEMALE.
 INTEGER(K4B):: WKI(NRDF)  !FOR EACH INDIVIDUAL, STORE THE BLOCK-ID IT BELONGS.
 REAL(DP):: DATALINE(NRDF,NEFF+1)
 INTEGER(K4B):: COUNT, NRBK !COUNTING VARIABLES.
 INTEGER(K4B):: i, j, k, st1 !DUMMY VARIABLES.

!*********
!PROCEDURE
!*********

!--------------------------------
!STEP 1.READ AND STORE DATA BLOCK
!--------------------------------
 WKI = 0; NRBK = 1

 OPEN (1,FILE='dat.txt',STATUS='OLD') !ATT. NAME OF THE INPUT DATA FILE. 

 COUNT = 1

 DO i = 1, NRDF

	!-----------------
	!DATA LINE READING
	!-----------------
  	 READ(1, *, IOSTAT=st1) DATALINE(i,:)

		!DATA READING CHECK BLOCK
		 IF (st1==0) THEN
	         IF(i == (COUNT*1000)) THEN
		         WRITE(*,'(i9,2x,a12)') i, 'RECORDS READ'
			     COUNT = COUNT + 1
			 ENDIF
		 ELSEIF (st1.EQ.-1) THEN !BY DEFAULT, st1 HAS VALUE -1 IF IT REACHES THE END OF FILE.
			 PRINT*, ''
			 PRINT*, 'BEWARE! UNEXPECTED END OF FILE. CHECK NRDF PARAMETER' 
			 STOP
		 ELSE
			 PRINT*, 'BEWARE! ERROR', st1,' IN LINE', i 
			 PRINT*, ''
			 STOP "FATAL ERROR: EXECUTION STOPPED"
		 END IF
		!---

 ENDDO
!---
!
!--------------------------
!STEP 2. FILL-UP WKI VECTOR
!--------------------------
 DO i = 1, NRDF

	 DAMI = 0

	!IF THE INDIVIDUAL IS A MALE, CYCLE.
	!ATT. "SEX" IT IS ASSUMED TO BE IN THE FIRST FIELD WITHIN THE DATA FILE.
	 IF (DATALINE(i,1).NE.2.D0) CYCLE

	!SET WKI FOR THE FIRST DAM WITHIN THE BLOCK:
	 IF (WKI(i).EQ.0) THEN
		 WKI(i) = NRBK
		 NRBK = NRBK + 1
	 ELSE
		 DAMI = 1
	 ENDIF

	!SEARCH FOR OFFSPRING, AND SET THE CORRESPONDING WKI
	!ATT. "IDM" AND "ID" ARE ASSUMED TO BE THE "NEFF" AND "NEFF-1" POSITIONS IN THE DATA LINE.
	 DO j = 1, NRDF
		 IF (DATALINE(j,NEFF).EQ. DATALINE(i,NEFF-1)) THEN
			 WKI(j) = WKI(i)
			 DAMI = 1
		 ELSE
			 CYCLE		 
		 ENDIF
	 ENDDO

	!CORRECT IF THE FEMALE IS NEITHER A DAM, NEITHER A DAM'S OFFSPRING
	 IF(DAMI.EQ.0) THEN
		 WKI(i)=0
		 NRBK = NRBK - 1
	 ENDIF

 ENDDO

 NBKS =	NRBK

!SEND UNLINKED INDIVIDUALS TO THE LAST BLOCK
 DO i = 1, NRDF
	 IF(WKI(i).EQ.0) WKI(i)=NBKS
 ENDDO
!---
!
!-------------------------
!STEP 3. PRINT-OUT RESULTS
!-------------------------
 OPEN (21,FILE='DAT1.sal',STATUS='REPLACE')
 
 ALLOCATE(BK(NBKS))
 
 DO j = 1, NBKS

	 BK(j)%FK=0



	 DO i = 1, NRDF
		 IF(WKI(i).EQ.j) THEN

			!----------------
			!REPARAMERIZATION
			!----------------
			!ATT. SYSTEM SHOULD BE POSITIVE DEFINITE. AVOID DEPENDENCIES SETTING SOME ROWS TO ZERO.
             DO k = 1, NEFF
                IF( REPR(k) .NE. 0) THEN
                     IF (DATALINE(i,k)==NLEV(k)) DATALINE(i,k) = 0 !SET TO ZERO THE LAST LEVEL OF THE EFFECT.  
                ENDIF
             ENDDO
			!---

			 WRITE(21,*) DATALINE(i,:)
			 BK(j)%FK = BK(j)%FK + 1

		 ELSE
			 CYCLE		 
		 ENDIF 		
	 ENDDO

 ENDDO
!---
 
 CLOSE (UNIT=1)
 CLOSE (UNIT=21)

 END SUBROUTINE FILE_PREP  
!***
!
!*****************************************************************************************
!INTERNAL SUBROUTINE SET_ADDRESS -------------------------------------- BY SML. 27/01/2010
!*****************************************************************************************
!
!SET (EFFECT, LEVEL,WITHIN-BLOCK INDIVIDUAL) ADDRESSES FOR EACH DATA FILE LINE.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!LOOPS OVER EFFECTS AND INDIVIDUALS WITHIN THE BLOCK, AND SET ADDRESSES GIVEN EFFECT TYPE. 
!*****************************************************************************************
!
!ATT. BEWARE! THIS SUBROUTINE MUST BE COMPILED WITHIN "RDBLK_GS v.1.0" PROGRAM.
!
 SUBROUTINE SET_ADDRESS
!
!---------------
!LOCAL VARIABLES 
!---------------
 INTEGER(K4B):: i, j

!*********
!PROCEDURE
!*********

 DO i = 1, NEFF !LOOP OVER EFFECTS

	 START = SUM(NLEV(1:i-1)) + 1

	 DO j = 1, BK(BLK)%FK !LOOP OVER INDIVIDUALS WITHIN BLOCKS
	
		 SELECT CASE(EFFT(i))
	
			!IF THE EFFECT IS A CATEGORICAL FACTOR:
			 CASE (CLSS)

				!IF THE INDIVIDUAL'S EFFECT IS SET TO ZERO:
				 IF(BK(BLK)%INDATA(i,j)==0) THEN
				 BK(BLK)%ADRESS(i,j) = 0
				 CYCLE
				 ENDIF
				!***

				 IF(BK(BLK)%INDATA(i,j)<0) THEN
				 BK(BLK)%ADRESS(i,j) = START + (-BK(BLK)%INDATA(i,j) - 1)
				 BK(BLK)%CVRW(i,j) = -1.D0

				 ELSE
				 BK(BLK)%ADRESS(i,j) = START + (BK(BLK)%INDATA(i,j) - 1)
				 BK(BLK)%CVRW(i,j) = 1.D0

				 ENDIF

			!IF THE EFFECT IS A COVARIATE:
			 CASE (COVR)
					
				 BK(BLK)%CVRW(i,j) = BK(BLK)%INDATA(i,j)	
				
				!IF THE COVARIATE IS NOT NESTED:
				 IF (NEST(i).eq.0) THEN

					 BK(BLK)%ADRESS(i,j) = START
				
				!IF THE COVARIATE IS NESTED:
				 ELSEIF (NEST(i).gt.0.AND.NEST(i).lt.NEFF) THEN
						
					 BK(BLK)%ADRESS(i,j) = START + (BK(BLK)%INDATA(NEST(i),j) - 1)
	
				 ELSE
				 PRINT*, 'ERROR IN NESTED COVARIATE. PLEASE, CHECK VARIABLE "NEST" SETTING'
				 STOP
	
				 ENDIF

			 CASE DEFAULT
			 PRINT*, 'UNIMPLEMENTED FIXED EFFECT', i
			 STOP
		
		 END SELECT

	 ENDDO

 ENDDO
!***
  
 END SUBROUTINE SET_ADDRESS  
!***
!
!*****************************************************************************************
!INTERNAL SUBROUTINE INVA---------------------------------------------- BY SML. 17/12/2008
!*****************************************************************************************
!
!RETURNS THE INVERSE ADDITIVE RELATIONSHIP MATRIX, INV-A, CONTRIBUTIONS TO THE MIXED MODEL
!EQUATIONS.
!***
!
!***VERSION 2.2.1***
!SAVE INVA v2.1 BUG. INBRED POPULATIONS. SPARSE STORAGE. DOUBLE PRECISION REAL VARIABLES.
!
!ALGORITHM DESCRIPTION********************************************************************
!BASED ON HENDERSON'S RULES [BIOMETRICS 1976, 32:69] AND MEUWISSEN & LUO'S WORK [GSE 1992,
!24:30]. CODE BASED ON SUBROUTINE "INBRED v2.0".
!*****************************************************************************************
!
!BEWARE! THIS SUBROUTINE MUST BE COMPILED WITHIN "RDBLK_GS" PROGRAM. 
!***
!
 SUBROUTINE INVA (ii)
!
!INTENT(IN): ii - ANIMAL EFFECT INDEX.
!***
!
!********************
!VARIABLE DECLARATION
!********************
 IMPLICIT NONE
!
!--------------
!DUMMY ARGUMENT
!--------------
 INTEGER(K4B), INTENT(IN):: ii
!---
!
!---------------
!LOCAL VARIABLES 
!---------------
 INTEGER(K4B):: IN, IS, ID, KS, KD !WORK VARIABLES.

	!MEUWISSEN AND LUO'S ALGORITHM VARIABLES
	 INTEGER(K4B):: POINT(1:N)      !POINTS TO THE NEXT ANCESTOR IN THE LINKED LIST.
	 INTEGER(K4B):: PED(1:3, 1:N)   !PEDIGREE
	 REAL(DP):: F(0:N) !INBREEDING COEFFICIENTS.
	 REAL(DP):: D(1:N) !MENDELIAN RESIDUAL VARIANCE CONTRIBUTION TO A-INV.
	 REAL(DP):: L(1:N) !FOR INDIVIDUAL i, ENTRIES (i,j) OF L MATRIX. 
	 REAL(DP):: R      !WORK VARIABLE. ACCUMULATE [0.5*L(i,k)] FOR k= {ANC}.
	 REAL(DP):: FI     !WORK VARIABLE. ACCUMULATE [L(i,j)^2*D(j,j)] FOR j= 1 TO i.

	!CONTRIBUTIONS VARIABLES
	 INTEGER(K4B):: ADDJJ, ADDKK, COUNT1, PFLAG
	 INTEGER(K4B):: P(3)
	 REAL(DP):: WORK(3) = (/1.D0,-0.5D0,-0.5D0/)

 INTEGER(K4B):: i, j, k,  c1, c2, jj, kk !DUMMY VARIABLES.
!---
!
!*********
!PROCEDURE
!*********

 POINT = 0		
  F(0) = -1.D0
	 L =  0.D0

!----------------------
!OPENING PEDIGREE FILES
!----------------------
 OPEN (2,file='ped.txt', STATUS='OLD')   !ATT. NAME OF PEDIGREE FILE. 
 OPEN (3,file='ERROR FILE.txt', STATUS='REPLACE')

 WRITE(3,'(a11)') '+++++++++++'
 WRITE(3,'(a11)') 'ERRORS FILE'
 WRITE(3,'(a11)') '+++++++++++'
 WRITE(3,*) ''
 WRITE(3,'(a24)') 'PEDIGREE INCONSISTENCIES'
 WRITE(3,'(a24)') '------------------------'
 WRITE(3,*) ''
 WRITE(3,'(a7,2x,a12)') '     ID', 'OBSERVATIONS'
 WRITE(3,'(a7,2x,a12)') '-------', '------------'
!---
!
!---------------------
!READING PEDIGREE FILE
!---------------------
 COUNT1 = 1
  PFLAG = 0

!-----------------------
!CYCLE AMONG INDIVIDUALS
!-----------------------
 DO i = 1, N

	!-------------------------------
	!PEDIGREE LINE READING & STORING
	!-------------------------------
	 READ(2, *) PED(:,i)

		!----------------
 		!PEDIGREE COUNTER
		!----------------
         IF(i.eq.(COUNT1*10000)) THEN
			 WRITE(*,'(i9,2x,a18)') i, 'TH INDIVIDUAL READ'
			 COUNT1 = COUNT1 + 1
		 ENDIF
		!---

		!------------------------------
		!CHECK PEDIGREE INCONSISTENCIES
		!------------------------------
		 IF (PED(2,i).GE.PED(1,i)) THEN
			 IF (PED(2,i).EQ.PED(1,i)) THEN
				 WRITE(3,'(i7,2x,a36)') PED(1,i), 'INDIVIDUAL AND ITS SIRE ID ARE EQUAL'
			 ELSE
			 	 WRITE(3,'(i7,2x,a36)') PED(1,i), 'INDIVIDUAL ID IS LOWER THAN ITS SIRE'
			 ENDIF
			 PED(2,i) = 0
			 PFLAG = 1
		 ELSEIF (PED(3,i).GE.PED(1,i)) THEN
			 IF (PED(3,i).EQ.PED(1,i)) THEN
				 WRITE(3,'(i7,2x,a35)') PED(1,i), 'INDIVIDUAL AND ITS DAM ID ARE EQUAL'
			 ELSE
			 	 WRITE(3,'(i7,2x,a35)') PED(1,i), 'INDIVIDUAL ID IS LOWER THAN ITS DAM'
			 ENDIF
			 PED(3,i) = 0
			 PFLAG = 1
		 ELSE
			 CONTINUE
		 ENDIF
		!---

	!----------------------------------
	!MEUWISSEN & LUO'S ALGORITHM KERNEL
	!----------------------------------
	 IN = PED(1,i)
	 IS = PED(2,i)
	 ID = PED(3,i)
	
	 PED(2,i) = MAX(IS,ID)
	 PED(3,i) = MIN(IS,ID)
	 
     D(i) = 0.5D0 - 0.25D0*(F(IS)+F(ID))

	!IF EITHER OF THE PARENTS IN UNKNOWN:
	 IF (IS.EQ.0 .OR. ID.EQ.0) THEN

	 	 F(i) = 0.D0 !INDIVIDUAL IS NOT INBRED

	!IF INDIVIDUALS i AND (i-1) ARE FULL SIBS:
	 ELSEIF (PED(2, i-1).EQ.PED(2,i) .AND. PED(3, i-1).EQ.PED(3,i)) THEN

		 F(i) = F(i-1) !BOTH HAS THE SAME INBREEDING COEFFICIENT

	 ELSE

		   FI = -1.D0
		 L(i) =  1.D0
		    j =  PED(1,i)

		 DO WHILE (j.NE.0)
		 
			  k = j
 			  R = 0.5D0*L(k)
			 KS = PED(2,k)
			 KD = PED(3,k)

			!IF ANCESTOR'S SIRE IS KNOWN:
			 IF(KS.GT.0) THEN

				 DO WHILE (POINT(k).GT.KS)
				 k = POINT(k)
				 ENDDO

				 L(KS) = L(KS) + R

				!IF ANCESTOR'S SIRE IS NOT NEXT IN THE ANCESTOR'S LIST: 
				 IF (KS.NE.POINT(k)) THEN
 				 POINT(KS) = POINT(k)
				 POINT(k)  = KS
				 ENDIF

				!IF ANCESTOR'S DAM IS KNOWN:
				 IF(KD.GT.0) THEN

					 DO WHILE (POINT(K).GT.KD)
					 k = POINT(k)
					 ENDDO

					 L(KD) = L(KD) + R

					!IF ANCESTOR'S DAM IS NOT NEXT IN THE ANCESTOR'S LIST:
					 IF (KD.NE.POINT(K)) THEN
 					 POINT(KD) = POINT(k)
					 POINT(k)  = KD
					 ENDIF

				 ENDIF
				!***

			 ENDIF
			!***

			 FI = FI + L(j)*L(j)*D(j)
			 L(j) = 0.D0
			 
			 k = j
			 j = POINT(j)
			 
			 POINT(k) = 0

		 ENDDO
		!***

		 F(i) = FI

	 ENDIF
	!---
	!
	!-------------------------------
	!ADDRESS SETTING & CONTRIBUTIONS
	!-------------------------------
	 P = (/IN, IS, ID/)
		 
	 DO c1 = 0, CORR(ii)-1
		 DO c2 = 0, CORR(ii)-1

			 DO jj = 1, 3
				 DO kk = 1, 3

					 IF (P(jj).NE.0 .AND. P(kk).NE.0) THEN

						!--------------
						!ADRESS SETTING
						!--------------
						 ADDJJ = SUM(NLEV(1:ii+c1-1)) + P(jj)
						 ADDKK = SUM(NLEV(1:ii+c2-1)) + P(kk)
						!---
						!
						!------------------------------
						!FULL IJA-STORE FOR AMAT MATRIX
						!------------------------------
						 IF(c1 .EQ. 0 .AND. c2 .EQ. 0) THEN

							 IF(ADDJJ.NE.0 .AND. ADDKK.NE.0) THEN 

								 AMAT%VALL1 = WORK(jj)*WORK(kk)*(1/D(i)) 

								 CALL SBSIJ(AMAT%N1,AMAT%IA1,AMAT%JA1,AMAT%A1,AMAT%IWK1,P(jj),P(kk),&
											AMAT%VALL1,MODE,SWITCH,NOUTT,IERR)

								!AN ERROR HAS OCCURRED. STOP PROCESS.
								 IF (IERR==701 .OR. IERR==702) STOP

							 ENDIF

						 ENDIF
						!--- 	
						!
						!--------------------------------------------------------------
						!ENTRIES CALCULATION: ONLY COMPUTE UPPER DIAGONAL CONTRIBUTIONS
						!--------------------------------------------------------------
						 IF(ADDJJ.LE.ADDKK) THEN 

							 VALL = INVG(ii,1+c1,1+c2)*WORK(jj)*WORK(kk)*(1/D(i)) 

							 CALL SBSIJ(NEQ,IA,JA,AVEC,IWORK,ADDJJ,ADDKK,VALL,MODE,SWITCH,NOUTT,IERR)

							!AN ERROR HAS OCCURRED. STOP PROCESS.
							 IF (IERR==701 .OR. IERR==702) STOP 

						 ENDIF
						!---

					 ENDIF

				 ENDDO
			 ENDDO

		 ENDDO
	 ENDDO
	!---

 ENDDO                                               
!---
!
!------------
!ERROR ADVICE
!------------
 IF(PFLAG==1) THEN
	 PRINT*, ''
	 PRINT*, 'BEWARE! PEDIGREE INCONSISTENCIES FOUND.'
	 PRINT*, 'SOME IDs HAS BEEN SET TO 0.'
	 PRINT*, 'CHECK ERROR FILE.'
 ELSE
 	 WRITE(3,*) ''
 	 WRITE(3,'(a35)') 'NO INCONSISTENCIES HAVE BEEN FOUND.'
 ENDIF
!---
!
!--------------------
!CLOSE OPRATIVE FILES
!--------------------
 CLOSE (UNIT=2)
 CLOSE (UNIT=3)
!
 END SUBROUTINE	INVA
!***

!*****************************************************************************************
!INTERNAL SUBROUTINE GIBBS--------------------------------------------- BY SML. 09/02/2010
!*****************************************************************************************
!
!WITHIN EACH ROUND, THIS ROUTINE SAMPLES FROM THE TRANSITION KERNEL OF A MARKOV CHAIN WITH
!THE MODEL'S PARAMETER VECTOR JOINT POSTERIOR DENSITY AS THE STATIONARY DISTRIBUTION, USI-
!NG A SINGLE-SITE SYSTEMATIC GIBBS SAMPLER ALGORITHM. THE TRANSITION KERNEL IS THE PRODUCT
!OF ALL FULL CONDITIONAL POSTERIORS DENSITIES, GIVEN THEY ARE KNOWN UP TO PROPORTIONALITY. 
!IT UPDATES THE WHOLE PARAMETER VECTOR SAMPLING FROM THE APPROPIATE DENSITIES.
!***
!
!***VERSION 2.0***
!MATERNAL ANIMAL MODEL WITH CORRELATED RESIDUALS. REQUIRE FULL IJA-STORE FOR THE NUMERATOR
!RELATIONSHIP MATRIX. USES A GRIDDY GIBBS ALGORITHM TO SAMPLE FROM THE CORRELATION PARAME-
!TER FULL CONDITIONAL DENSITY.
!
!ALGORITHM DESCRIPTION********************************************************************
!THEORY BASED ON SORENSEN & GIANOLA BOOK (2002). CODE ADPATED FROM "GS_FCG D" F77 PROGRAM,
!WRITTEN BY PROFESSOR R.J.C. CANTET, AND FROM "GIBBS1.F90" PROGRAM, WRITTEN BY PROFESSOR I.
!MISTZAL. GRIDDY GIBBS ALGORITHM AS DESCRIBED BY...
!UPDATING ORDER: S2ER, TITA, pPAR, G-MATRICES. REQUIERE	3 WITHIN-BLOCK LOOPS.
!*****************************************************************************************
!
!BEWARE! THIS SUBROUTINE MUST BE COMPILED WITHIN "RDBLK_GS v1.0" PROGRAM.
!***
!
 SUBROUTINE GIBBS(ROUND)
!
!INTENT(IN): ROUND - ITERATION NUMBER.
!***

!********************
!VARIABLE DECLARATION
!********************
 IMPLICIT NONE
!
!---------------
!DUMMY ARGUMENTS
!---------------
 INTEGER(K4B), INTENT(IN):: ROUND
!
!---------------
!LOCAL VARIABLES 
!---------------
!LOCATION PARAMETERS SAMPLING
 INTEGER(K4B):: RNDP(NEFF)
 REAL(DP):: DIAG
 REAL(DP):: TEMPOR(NEQ), UPTITA(NEQ)
 INTEGER(K4B):: OFFSET(NEFF,MAXX,2)
 INTEGER(K4B), ALLOCATABLE:: IA2(:), JA2(:), IWORK2(:)
 REAL(DP), ALLOCATABLE:: XY1(:), AVEC2(:)
 REAL(DP):: G_BLOCK(NEFF,MAXX)

!ERROR VARIANCE SAMPLING
 REAL(DP):: SSR, VE2, S2E, NEWR

!CORRELATION PARAMETER SAMPLING
 INTEGER(K4B):: IDX, INDI
 REAL(DP):: SSR1(NGRID), DET1(NGRID), DETR1(NGRID), CDF(NGRID)
 REAL(DP):: MAXV, WJ, WJ1, UNI
     
!(CO)VARIANCE COMPONENTS SAMPLING
 INTEGER(K4B):: M1, M2, N1, N2, G1, G2
 REAL(DP):: QFORM(NEFF, MAXX, MAXX)
 REAL(DP):: VU2
 INTEGER(K4B):: RANK
 REAL(DP), ALLOCATABLE:: D1(:), NEWG(:)
 
!DUMMY VARIABLES.
 INTEGER(K4B):: i, j, l, m, ii, jj, t
!---
!
!*********
!PROCEDURE
!*********
!
!---------------------------
!PRINT PARAMETER MCMC CHAINS
!---------------------------
!ATT. BEWARE OUTPUT FORMATS.
 WRITE(25,'(100(f11.5,2x))') S2ER, pPAR, G(5,1,1), G(6,1,1), G(6,2,1), G(6,2,2)
!---
!
!--------------------------------------------
!ALLOCATE BOUNDS & INITIALIZE LOCAL VARIABLES
!--------------------------------------------
!LOCATION PARAMETER SAMPLING
 ALLOCATE (XY1(NEQ), IA2(NEQ+1), JA2(NZ), IWORK2(NZ), AVEC2(NZ))
 CALL SBINI(NEQ,NZ,IA2,JA2,AVEC2,IWORK2)
 XY1 = 0.D0

!ERROR VARIANCE SAMPLING
 SSR = 0.D0

!CORRELATION PARAMETER SAMPLING
!ATT. RESET GLOBAL GRID
 GRIDD(1) = MIN_VAL 
 DO i = 2, NGRID
     GRIDD(i) = GRIDD(i-1) + INCR
 ENDDO    
!---
!
!-----------------------------------------------------------------
!LOOP 1 - SET-UP UPDATED MIXED MODEL EQUATIONS & COMPUTE RESIDUALS
!-----------------------------------------------------------------

!CYCLE AMONG BLOCKS
 DO BLK = 1, NBKS

	!REINITIALIZE WITHIN-BLOCK OPERATIVE VARIABLES
	 BK(BLK)%R = 0; BK(BLK)%RINV = 0; BK(BLK)%RES = 0
	!---
	!
	!ONLY FOR THE FIRST ROUND, ADJUST ADDRESSES GIVEN COLLAPSED SYSTEM COLUMN INDICATOR
	 IF(ROUND==1) THEN	
		 DO i = 1, NEFF
			 DO k = 1, BK(BLK)%FK
				 IF (BK(BLK)%ADRESS(i,k)==0) CYCLE !IF IT IS A DUMMY ADDRESS, CYCLE.
				 BK(BLK)%ADRESS(i,k) = BK(BLK)%ADRESS(i,k) - DCOL(BK(BLK)%ADRESS(i,k))
			 ENDDO
		 ENDDO
	 ENDIF
	!---

		!-----------------------------
		!1. COMPUTE & INVERT Rk-MATRIX
		!-----------------------------
		 DO i = 1, BK(BLK)%FK
			 BK(BLK)%R(i,i) = 1.D0
			 DO j = i+1, BK(BLK)%FK
				 IF (BK(BLK)%INDATA(NEFF,j).EQ.BK(BLK)%INDATA(NEFF-1,i)) THEN
				 BK(BLK)%R(i,j) = pPAR
				 BK(BLK)%R(j,i) = BK(BLK)%R(i,j)
				 ENDIF
			 ENDDO
		 ENDDO
		 BK(BLK)%RINV = BK(BLK)%R
  		 IF(BLK.NE.NBKS) CALL GINV(BK(BLK)%RINV,BK(BLK)%FK,BK(BLK)%FK,1.D-12,RANK)
		!---
		!
		!-----------------------------------------------
		!1.1. COMPUTE IJA-FORMAT W'*RINV*W AND W'*RINV*Y
		!-----------------------------------------------
		 DO i = 1, NEFF
			 DO k = 1, BK(BLK)%FK
				 DO j = 1, NEFF

					!EXCEPT FOR THE LAST BLOCK
					 IF(BLK.NE.NBKS) THEN

						 DO l = 1, BK(BLK)%FK
							 IF( BK(BLK)%ADRESS(i,k).LE.BK(BLK)%ADRESS(j,l) .AND.&
								 BK(BLK)%ADRESS(i,k).NE.0 .AND. BK(BLK)%ADRESS(j,l).NE.0) THEN
								 VALL = BK(BLK)%CVRW(i,k)*BK(BLK)%CVRW(j,l)*BK(BLK)%RINV(k,l) 
								 CALL SBSIJ(NEQ,IA2,JA2,AVEC2,IWORK2,BK(BLK)%ADRESS(i,k),&
											BK(BLK)%ADRESS(j,l),VALL,MODE,SWITCH,NOUTT,IERR)
								 IF (IERR==701 .OR. IERR==702) STOP 
							 ENDIF
						 ENDDO
					 ELSE
						 IF( BK(BLK)%ADRESS(i,k).LE.BK(BLK)%ADRESS(j,k) .AND.&
							 BK(BLK)%ADRESS(i,k).NE.0 .AND. BK(BLK)%ADRESS(j,k).NE.0) THEN
							 VALL = BK(BLK)%CVRW(i,k)*BK(BLK)%CVRW(j,k)*BK(BLK)%RINV(k,k) 
							 CALL SBSIJ(NEQ,IA2,JA2,AVEC2,IWORK2,BK(BLK)%ADRESS(i,k),&
										BK(BLK)%ADRESS(j,k),VALL,MODE,SWITCH,NOUTT,IERR)
							 IF (IERR==701 .OR. IERR==702) STOP 
						 ENDIF
					 ENDIF
				 ENDDO
				 IF(BLK.NE.NBKS) THEN
					 DO l = 1, BK(BLK)%FK
						 IF (BK(BLK)%ADRESS(i,k)==0) CYCLE
						 XY1(BK(BLK)%ADRESS(i,k)) = XY1(BK(BLK)%ADRESS(i,k)) + BK(BLK)%Y(l)*&
													BK(BLK)%CVRW(i,k)*BK(BLK)%RINV(k,l)
					 ENDDO
				 ELSE
					 IF (BK(BLK)%ADRESS(i,k)==0) CYCLE
					 XY1(BK(BLK)%ADRESS(i,k)) = XY1(BK(BLK)%ADRESS(i,k)) + BK(BLK)%Y(k)*&
												BK(BLK)%CVRW(i,k)*BK(BLK)%RINV(k,k)
				 ENDIF
			 ENDDO
		 ENDDO 
		!---
		!
		!-------------------------------------------------
		!1.2. COMPUTE RESIDUALS & RESIDUALS SUM OF SQUARES
		!-------------------------------------------------
		 DO k = 1, BK(BLK)%FK
			 BK(BLK)%RES(k) = BK(BLK)%Y(k) 
			 DO i = 1, NEFF
				 IF (BK(BLK)%ADRESS(i,k)==0) CYCLE
				 BK(BLK)%RES(k) = BK(BLK)%RES(k) - BK(BLK)%CVRW(i,k)*TITA(BK(BLK)%ADRESS(i,k))
			 ENDDO    
		 ENDDO
		 DO i = 1, BK(BLK)%FK
			 DO j = 1, BK(BLK)%FK
				 SSR = SSR + BK(BLK)%RES(i)*BK(BLK)%RES(j)*BK(BLK)%RINV(i,j)
			 ENDDO
		 ENDDO
		!---

 ENDDO
!---
!
!BUILD SPARSE-STORAGE FINAL DATA STRUCTURE
 CALL SBEND(NEQ,NZ,IA2,JA2,AVEC2,IWORK2)
!---
!
!-----------------------
!ERROR VARIANCE SAMPLING
!-----------------------
!SET SCALED INVERTED CHI-SQUARE PARAMETERS
!DEGREES OF BELIEF 
 VE2 = 1.D0*(NRDF + VE1)

!SCALE PARAMETER 
!S2E = (SSR + (S2E0 * VE1)) / VE2
 S2E = (SSR + (S2E0 * (VE1+1)))

!SAMPLE FROM A SCALED INVERTED CHI-SQUARE DENSITY
!CALL CHIN(S2E, VE2, SEED, NEWR)
 NEWR = ICHSQ(VE2, S2E, SEED)

!TEST FOR BIG SAMPLE VALUE
 IF(NEWR > 200000.D0) STOP " BEWARE! S2E0 SAMPLE VALUE TOO BIG!"

!UPDATE ERROR VARIANCE
 S2ER = NEWR
!---
!
!----------------------------
!LOCATION PARAMETERS SAMPLING
!----------------------------
 UPTITA = 0.D0
 TEMPOR = XY1 !INITIALIZE "TEMPOR" WITH XY1 VECTOR.

!SET OFFSETS AND POINTERS FOR COVARIANCE MATRICES. 
 CALL SET_RANDOM_BLOCKS (RNDP,OFFSET,G_BLOCK)

!SAMPLING LOOP.
 DO ii = 1, NEFF 
      DO jj = 1, NLEV1(ii)

         i = SUM(NLEV1(1:ii-1)) + jj

		 DIAG = 0

		!CONTRIBUTIONS DUE TO LOWER-DIAGONAL W'*RINV*W MATRIX:
		 TEMPOR(i) = TEMPOR(i) - UPTITA(i)

		!CONTRIBUTIONS DUE TO UPPER-DIAGONAL W'*RINV*W MATRIX:
		 DO j = IA2(i), IA2(i+1)-1
			 IF (i .NE. JA2(j)) THEN
				 TEMPOR(i) =  TEMPOR(i) - AVEC2(j)*TITA(JA2(j))
			 ELSE
				 DIAG = AVEC2(j)
			 ENDIF
		 ENDDO

		!CONTRIBUTIONS DUE TO INVERSE COVARIANCE MATRICES:
		 SELECT CASE(TYPO(ii))
	
 			!IF IT IS A FIXED EFFECT:
			 CASE (FIXX)

 				!ADD A VERY SMALL QUANTITY TO THE DIAGONAL ENTRY TIMES UPDATED R:
				 DIAG = DIAG + (S2ER/10000000.D0)


			!IF IT IS A RANDOM iid EFFECT:
			 CASE (RND1)
			
				 DIAG = DIAG + (S2ER*INVG(ii,1,1))


			!IF IT IS A RANDOM ANIMAL EFFECT:
			 CASE (RND2)

				 l = RNDP(ii)

				 IF (l .NE. 0) THEN

					 DO m = 1, CORR(l)

		 				 DO j = AMAT%IA1(i-OFFSET(ii,m,1)),&
							    AMAT%IA1(i-OFFSET(ii,m,1)+1)-1

							 IF (i .NE. AMAT%JA1(j) + OFFSET(ii,m,2)) THEN

								 TEMPOR(i) =  TEMPOR(i) - AMAT%A1(j)*(S2ER*G_BLOCK(ii,m))*&
											  TITA(AMAT%JA1(j) + OFFSET(ii,m,2))
							 ELSE

								 DIAG = DIAG + (S2ER*G_BLOCK(ii,m))*AMAT%A1(j)
					
							 ENDIF

						 ENDDO

			         ENDDO

				 ENDIF    

			!ELSE:
			 CASE DEFAULT

				 PRINT*, 'UNIMPLEMENTED RANDOM EFFECT', TYPO(ii)
				 STOP
		
		 END SELECT
		!***

		!COMPUTE FULL CONDITIONAL POSTERIOR MEAN
		 TITA(i) = TEMPOR(i)/DIAG

		!SAMPLE TITA FROM A GAUSSIAN DENSITY
		 TITA(i) = TITA(i) + (1.D0 * GASDEV(SEED)) / DSQRT(DIAG*(1.D0/S2ER))

		!COMPUTE W'*RINV*W(i,1:NEQ)*UPDATED_TITA(1:NEQ)
		 DO j = IA2(i), IA2(i+1)-1
			 UPTITA(JA2(j)) = UPTITA(JA2(j)) + AVEC2(j)*TITA(i)
		 ENDDO

	 ENDDO

 ENDDO
!***
!
!-------------------------------------
!LOOP 2 - SEARCH FOR LOCAL GRID VALUES
!-------------------------------------
!
 SSR1 = 0.D0; DET1 = 0.D0; DETR1 = 0.D0;

!CYCLE AMONG BLOCKS
 DO BLK = 1, NBKS

	!------------------------------
	!CYCLE ALONG GLOBAL GRID VALUES
	!------------------------------
	 DO m = 1, NGRID

	 	!REINITIALIZE WORKING VARIABLES
		 BK(BLK)%R = 0; BK(BLK)%RINV = 0; BK(BLK)%RES = 0

		!COMPUTE & INVERT Rk-MATRIX FOR EACH pPAR (GLOBAL GRID)
		 DO i = 1, BK(BLK)%FK
			 BK(BLK)%R(i,i) = 1.D0
			 DO j = i+1, BK(BLK)%FK
				 IF (BK(BLK)%INDATA(NEFF,j).EQ.BK(BLK)%INDATA(NEFF-1,i)) THEN
				 BK(BLK)%R(i,j) = GRIDD(m)
				 BK(BLK)%R(j,i) = BK(BLK)%R(i,j)
				 ENDIF
			 ENDDO
		 ENDDO
		 BK(BLK)%RINV = BK(BLK)%R
		 IF(BLK.NE.NBKS) THEN
			 CALL GINV(BK(BLK)%RINV,BK(BLK)%FK,BK(BLK)%FK,1.D-12,RANK,DET1(m))
				 IF(RANK .LT. BK(BLK)%FK) THEN
					 PRINT*, ''
					 PRINT*, '-----------'
					 WRITE(*,'(a30,i3,a8,i1,a27)') 'FATAL ERROR: RINV FOR FAMILY ', BLK,&
												   ', GRIDD ', m,', IS NOT POSITIVE DEFINITE.' 
					 PRINT*, '-----------'
					 PRINT*, ''
					  STOP
				 ENDIF
			 DET1(m) = LOG(DET1(m))
		 ENDIF

		!COMPUTE RESIDUALS & ACCUMULATE RESIDUALS SUM OF SQUARES FOR EACH pPAR (GLOBAL GRID)
		 DO k = 1, BK(BLK)%FK
			 BK(BLK)%RES(k) = BK(BLK)%Y(k)
			 DO i = 1, NEFF
				 IF (BK(BLK)%ADRESS(i,k)==0) CYCLE
				 BK(BLK)%RES(k) = BK(BLK)%RES(k) - BK(BLK)%CVRW(i,k)*TITA(BK(BLK)%ADRESS(i,k))
			 ENDDO    
		 ENDDO
		 DO i = 1, BK(BLK)%FK
			 DO j = 1, BK(BLK)%FK
				 SSR1(m) = SSR1(m) + BK(BLK)%RES(i)*BK(BLK)%RES(j)*BK(BLK)%RINV(i,j)
			 ENDDO
		 ENDDO

		!SUM UP R-MATRIX LOG-DETERMINANTS FOR EACH pPAR (GLOBAL GRID)
		 IF(BLK.EQ.NBKS) DET1(m) = 0.D0
		 DETR1(m) = DETR1(m) + DET1(m)

	 ENDDO
	!---

 ENDDO
!----

!EVALUATE LOG-FUNCTION
 DO j = 1, NGRID
	 SSR1(j) = -0.5D0*((SSR1(j)/S2ER) + DETR1(j))
 ENDDO
!---

!SEARCH MAX{LOG-FUNCTION}
 IDX = 0
 DO j = 1, NGRID
	 IF(SSR1(j) == MAXVAL(SSR1)) THEN
		 IDX=j
		 EXIT
	 ENDIF
 ENDDO
 MAXV = GRIDD(IDX)
!---
!
!RESET LOCAL GRIDD
 DO j = 1, NGRID
      GRIDD(j) = MAXV+((j-MID_VAL)*TUN)
 ENDDO
!
!RESET WORKING VARIABLES
 SSR1 = 0.D0; DET1 = 0.D0; DETR1 = 0.D0;
!
!------------------------------------------------------------
!LOOP 3 - COMPUTE RESIDUALS & LOG-DETERMINANTS FOR LOCAL GRID
!------------------------------------------------------------
!
!CYCLE AMONG BLOCKS
 DO BLK = 1, NBKS

	!-----------------------------
	!CYCLE ALONG LOCAL GRID VALUES
	!-----------------------------
	 DO m = 1, NGRID

	 	!REINITIALIZE WORKING VARIABLES
		 BK(BLK)%R = 0; BK(BLK)%RINV = 0; BK(BLK)%RES = 0

		!COMPUTE & INVERT Rk-MATRIX FOR EACH pPAR (LOCAL GRID)
		 DO i = 1, BK(BLK)%FK
			 BK(BLK)%R(i,i) = 1.D0
			 DO j = i+1, BK(BLK)%FK
				 IF (BK(BLK)%INDATA(NEFF,j).EQ.BK(BLK)%INDATA(NEFF-1,i)) THEN
				 BK(BLK)%R(i,j) = GRIDD(m)
				 BK(BLK)%R(j,i) = BK(BLK)%R(i,j)
				 ENDIF
			 ENDDO
		 ENDDO
		 BK(BLK)%RINV = BK(BLK)%R

		 IF(BLK.NE.NBKS) THEN 
			 CALL GINV(BK(BLK)%RINV,BK(BLK)%FK,BK(BLK)%FK,1.D-12,RANK,DET1(m))
				 IF(RANK .LT. BK(BLK)%FK) THEN
					 PRINT*, ''
					 PRINT*, '-----------'
					 WRITE(*,'(a30,i3,a8,i1,a27)') 'FATAL ERROR: RINV FOR FAMILY ', BLK,&
												   ', GRIDD ', m,', IS NOT POSITIVE DEFINITE.' 
					 PRINT*, '-----------'
					 PRINT*, ''
					  STOP
				 ENDIF
			 DET1(m) = LOG(DET1(m))
		 ENDIF

		!COMPUTE RESIDUALS & ACCUMULATE RESIDUALS SUM OF SQUARES FOR EACH pPAR (LOCAL GRID)
		 DO k = 1, BK(BLK)%FK
			 BK(BLK)%RES(k) = BK(BLK)%Y(k)
			 DO i = 1, NEFF
				 IF (BK(BLK)%ADRESS(i,k)==0) CYCLE
				 BK(BLK)%RES(k) = BK(BLK)%RES(k) - BK(BLK)%CVRW(i,k)*TITA(BK(BLK)%ADRESS(i,k))
			 ENDDO    
		 ENDDO
		 DO i = 1, BK(BLK)%FK
			 DO j = 1, BK(BLK)%FK
				 SSR1(m) = SSR1(m) + BK(BLK)%RES(i)*BK(BLK)%RES(j)*BK(BLK)%RINV(i,j)
			 ENDDO
		 ENDDO

		!SUM UP R-MATRIX LOG-DETERMINANTS FOR EACH pPAR (LOCAL GRID)
		 IF(BLK.EQ.NBKS) DET1(m) = 0.D0
		 DETR1(m) = DETR1(m) + DET1(m)

	 ENDDO
	!---
		 
 ENDDO
!---
!
!------------------------------
!CORRELATION PARAMETER SAMPLING
!------------------------------
!FOLLOWING ALGORITHM BY LIESENFELD, R. (2008): FINANCIAL DATA ANALYSIS, AND DR. LUC BAUWENS ADVICE.
!LOGARITHM OF THE SUM IS FOLLOWING THE ALGORITHM BY PRIMEAUX (2005).
 WJ1=0.D0
  WJ=0.D0

!STEP 1. COMPUTE w(j) FROM THE p-PARAMETER FULL CONDITIONAL DENSITY KERNEL
 DO j = 1, NGRID

	 SSR1(j) = -0.5D0*((SSR1(j)/S2ER) + DETR1(j))

	!ACCUMULATOR
	 IF (j==1) THEN
		 WJ = SSR1(j)
		 CYCLE
	 ENDIF
	 WJ1 = SSR1(j) - WJ 
	 WJ1 = EXP(WJ1) + 1.D0
	 WJ1 = LOG(WJ1)
	  WJ = WJ1 + WJ

 ENDDO
!---
!
!STEP 2. APPROXIMATE p-PARAMETER FULL CONDITIONAL CDF
 CDF = 0.D0
 DO j = 1, NGRID
	 DO t = 1, j
		 CDF(j) = CDF(j) + EXP(SSR1(t) - WJ)
	 ENDDO
 ENDDO
!---
!
!STEP 3. SAMPLE A UNIFORM[MIN(CDF), MAX(CDF)] DEVIATE
 INDI = 0
 t=0 !COUNTS THE NUMBER OF RANDOM UNIFORM DEVIATES SAMPLED
 DO WHILE (INDI == 0)
	 t=t+1
	 UNI = RAN(SEED)
	 IF(UNI .GE. CDF(1) .AND. UNI .LE. CDF(NGRID)) INDI = 1

	!TEST
	 IF(t==20) THEN
		 PRINT*, ''
		 PRINT*, 'BEWARE! TRUNCATED CDF AT CYCLE: ', ROUND
		 PRINT*, 'CDF BETWEEN ', CDF(1), ' AND ', CDF(NGRID)
		 PRINT*, ''
	 ENDIF
	 IF(t==1000) STOP "FATAL ERROR: PROGRAM TERMINATED"
	!END_TEST

 ENDDO
!---
!
!STEP 4. SEARCH FOR INDEX 'IDX' SUCH THAT CDF(IDX) < UNI < CDF(IDX+1)
 IDX = 0
 DO j = 1, NGRID
	 IF(CDF(j) < UNI) THEN
		 IDX=j
	 ENDIF
 ENDDO
!---
!
!STEP 5. UPDATE pPAR BY LINEAR INTERPOLATION BETWEEN GRIDD(IDX) & GRIDD(IDX+1)
 pPAR = GRIDD(IDX) + (GRIDD(IDX+1) - GRIDD(IDX)) * (UNI - CDF(IDX))/(CDF(IDX+1) - CDF(IDX))

!TEST FOR SAMPLE OUT OF PARAMETER SPACE
 IF(pPAR < -1.D0 .OR.  pPAR > 1.D0) STOP " BEWARE! p_PAR SAMPLE OUT OF PARAMETRIC SPACE!"
!---
!
!----------------------------
!VARIANCE COMPONENTS SAMPLING
!----------------------------
 QFORM = 0.D0

!COMPUTE QUADRATIC AND BILINEAR FORMS 
 DO i = 1, NEFF

	!IF IT IS A RANDOM EFFECT, NOT CORRELATED WITH A PREVIOUS ONE:
	 IF (CORR(i) .GT. 0) THEN

		!SET j=0 FOR DIRECT EFFECT AND j=1 FOR MATERNAL
	     DO j = 0, CORR(i) - 1

			!SET k=0 FOR DIRECT EFFECT AND k=1 FOR MATERNAL
			 DO k = 0, CORR(i) - 1

				!SET ADDRESS WHERE i-EFFECT START & END.
				 M1 = SUM(NLEV1(1:(i+j)-1)) + 1; M2 = SUM(NLEV1(1:(i+j)-1)) + NLEV1(i)
				 N1 = SUM(NLEV1(1:(i+k)-1)) + 1; N2 = SUM(NLEV1(1:(i+k)-1)) + NLEV1(i)

				!SET G1=1 FOR DIRECT EFFECT AND G1=2 FOR MATERNAL
				 G1= j + 1

				!SET G2=1 FOR DIRECT EFFECT AND G2=2 FOR MATERNAL
				 G2= k + 1

				!COMPUTE ONLY BELOW DIAGONAL BILINEAR FORMS
				 IF (G1 .LE. G2) THEN

					!FOR IID RANDOM EFFECTS
					 IF(TYPO(i) .EQ. RND1) THEN

						!COMPUTE QUADRATIC FORM
						 QFORM(i,G1,G2) = DOT_PRODUCT(TITA(M1:M2),TITA(N1:N2))

					!FOR RANDOM ANIMAL EFFECTS
					 ELSEIF(TYPO(i) .EQ. RND2) THEN

						!COMPUTE QUADRATIC FORMS
						 QFORM(i,G1,G2) = QFORM_IJA(TITA(M1:M2),AMAT,TITA(N1:N2))

   						!COPY ABOVE DIAGONAL BILINEAR FORMS
						 QFORM(i,G2,G1) = QFORM(i,G1,G2)	

					 ELSE

						STOP "BEWARE! CHECK <<TYPO>> AND <<CORR>> DECLARATION"

					 ENDIF

				 ENDIF

			 ENDDO

		 ENDDO
		!***

		!SET SCALED INVERTED WISHART PARAMETERS, SAMPLE AND UPDATE G.
		 j = CORR(i)

	     IF (j .GT. 0) THEN

			!FOR IID RANDOM EFFECTS
			 IF(TYPO(i) .EQ. RND1) THEN

				!SET SCALED INVERTED CHI-SQUARE PARAMETERS
				!DEGREES OF BELIEF 
				 VE2 = 1.D0*(NLEV1(i) + VEM1)

				!SCALE PARAMETER 
!				 S2E = (QFORM(i,1,1) + (G0(i,1,1) * VEM1)) / VE2
				 S2E = (QFORM(i,1,1) + (G0(i,1,1) * (VEM1+2)))

				!SAMPLE FROM A SCALED INVERTED CHI-SQUARE DENSITY
			 	 ALLOCATE(NEWG(1))

!				 CALL CHIN(S2E, VE2, SEED, NEWG(1))
                 NEWG(1) = ICHSQ(VE2, S2E, SEED)

				!UPDATE IID RANDOM EFFECT VARIANCE
				 G(i,1,1) = NEWG(1)

			 	 DEALLOCATE(NEWG)


			!FOR RANDOM ANIMAL EFFECTS
			 ELSEIF(TYPO(i) .EQ. RND2) THEN

				!COMPUTE SCALE MATRIX:
				!ADD A PRIORI SCALE TO THE QFORM
				 QFORM(i,1:j,1:j) = QFORM(i,1:j,1:j) + G0(i,1:j,1:j)*(VU1-CORR(i) - 1)

				!INVERT SCALE MATRIX
				 CALL GINV(QFORM(i,1:j,1:j),j,MAXX,1.D-12,RANK)

					!SECURITY DEVICE: CHECK OUT IF SCALE MATRIX IS POSITIVE DEFINITE.
					 IF(RANK .LT. j) THEN
						 PRINT*, ''
						 PRINT*, '-----------'
						 WRITE(*,'(a27,i1,a27)') 'FATAL ERROR: MATRIX QFORM(', i, ') IS&
												  &NOT POSITIVE DEFINITE.' 
						 PRINT*, '-----------'
						 PRINT*, ''
						  STOP
					 ENDIF

				!PERFORM CHOLESKY DECOMPOSITION
			 	 ALLOCATE(D1(j*(j+1)/2), NEWG(j*(j+1)/2))

				 CALL CHOLESKY (QFORM(i,1:j,1:j), j, 1.D-12, RANK, D1)

				!SET DEGREES OF BELIEF:
				 VU2 = VU1 + AMAT%N1

				!SAMPLE FROM A SCALED WISHART DENSITY
			 	 CALL WSHRT(D1, VU2, j, SIZE(NEWG), NEWG, SEED)

				!SET UPDATED G(i,1:j,1:j) FINAL STRUCTURE
				 m = 0

				!FILL ABOVE DIAGONAL ENTRIES 
				 DO jj = 1, j
					 DO ii = 1, jj
						m = m + 1
						G(i,jj,ii) = NEWG(m)
					 ENDDO
				 ENDDO

				!FILL BELOW DIAGONAL ENTRIES 
				 DO jj = 1, j
					 DO ii = 1, jj - 1
						G(i,ii,jj) = G(i,jj,ii)
					 ENDDO
				 ENDDO

				!FINALLY, INVERT TO OBTAIN A SCALED INVERSE WISHART REALIZATION
				 CALL GINV(G(i,1:j,1:j),j,MAXX,1.D-12,RANK)

					!SECURITY DEVICE: CHECK OUT IF SCALE MATRIX IS POSITIVE DEFINITE.
					 IF(RANK .LT. j) THEN
						 PRINT*, ''
						 PRINT*, '-----------'
						 WRITE(*,'(a26,i2,a27)') 'FATAL ERROR: MATRIX NEWG(', i, ') IS &
												  &NOT POSITIVE DEFINITE.' 
						 WRITE(*,'(a13,i7)') 'ROUND:', ROUND
						 PRINT*, '-----------'
						 PRINT*, ''
						  STOP
					 ENDIF

				!DEALLOCATE WORKING VECTORS.
				 DEALLOCATE(D1, NEWG)

			 ELSE

				STOP "BEWARE! CHECK <<TYPO>> AND <<CORR>> DECLARATION"

			 ENDIF

	     ENDIF
		!***

	 ENDIF

 ENDDO

!DEALLOCATE WORK VARIABLES
 DEALLOCATE(IA2, JA2, IWORK2, XY1, AVEC2)
!---

 END SUBROUTINE GIBBS
!***
!
!*****************************************************************************************
!INTERNAL SUBROUTINE SET_RANDOM_BLOCKS -------------------------------- BY SML. 21/05/2009
!*****************************************************************************************
!
!SET POINTERS AND VARIANCE COMPONENTS ASSOCIATED TO EACH BLOCK OF A COVARIANCE MATRIX WITH
!KRONECKER STRUCTURE, SUCH AS THE RELATIONSHIP MATRIX IN A CORRELATED TRAIT MODEL. BLOCK j 
!STANDS FOR AN EQUATION OF THE jth-CORRELATED COMPONENT IN THIS CASE.
!***
!
!ALGORITHM DESCRIPTION********************************************************************
!PROFESSOR IGNACY MISTZAL'S CODE, SLIGHTLY ADAPTED TO ONE-EQUATION BLOCKS. 
!*****************************************************************************************

 SUBROUTINE SET_RANDOM_BLOCKS (RNDP,OFFSET,G_BLOCK)

!INTENT(INOUT)::        RNDP(NEFF) - POSITION OF THE FIRST EFFECT OF THE CORRELATED BLOCK. 
!INTENT(OUT)::  G_BLOCK(NEFF,MAXX) - RETURNS THE CORRESPONDING VARIANCE COMPONENT.
!INTENT(OUT):: OFFSET(NEFF,MAXX,2) - RETURNS ADDRESSES OF THE BLOCK TO BE COMPUTED.

!***********
!DECLARATION
!***********

!DUMMY ARGUMENTS
 INTEGER(K4B), INTENT(INOUT):: RNDP(:)
 INTEGER(K4B), INTENT(OUT):: OFFSET(NEFF,MAXX,2)
 REAL(DP), INTENT(OUT):: G_BLOCK(NEFF,MAXX)

!LOCAL VARIABLES 
 INTEGER(K4B):: i, j, k

!*********
!PROCEDURE
!*********
  
!INITIALIZATION.
 RNDP = 0; OFFSET = 0; G_BLOCK = 0         
  
!SETTING RNDP VECTOR
 i=1
 DO WHILE(i .LE. NEFF)
	 IF(CORR(i) .GT. 0) THEN
		 DO j = 0, CORR(i)-1
			 RNDP(i+j) = i
		 ENDDO
		 i = i + CORR(i)  
     ELSE
		 i = i + 1
     ENDIF
 ENDDO
!***

!SETTING OFFSETS & G_BLOCK
 DO i = 1, NEFF
	 k = RNDP(i)
     IF(k /= 0) THEN
		 DO j = 1, CORR(k)
			 OFFSET(i,j,1) = SUM(NLEV1(1:i-1))     !ROW POINTER.
			 OFFSET(i,j,2) = SUM(NLEV1(1:(k+j-2))) !COLUMN POINTER.
              G_BLOCK(i,j) = INVG(k,(i-k+1),j)	   !INVG MATRIX ENTRY.
		 ENDDO
	 ENDIF
 ENDDO
!***

 END SUBROUTINE SET_RANDOM_BLOCKS  
!***
!
!*****************************************************************************************
!FUNCTION QFORM_IJA --------------------------------------------------- BY SML. 19/06/2009
!*****************************************************************************************
!
!IGNACYS MISTZAL'S CODE. COMPUTE A BILINEAR FORM WHEN THE ASSOCIATED MATRIX IS SQUARE, SY-
!MMETRIC AND FULL IJA-STORED.
!***
!
 FUNCTION QFORM_IJA(X,A,Y) RESULT(Q)

!***********
!DECLARATION
!***********
 IMPLICIT NONE

!DUMMY ARGUMENT
 TYPE(SPARSE_MATRIX):: A
 REAL(DP):: X(:), Y(:)
!
!LOCAL VARIABLES
 REAL(DP):: Q
 INTEGER(K4B):: i, j, k
!***
!
!*********
!PROCEDURE
!*********
!
!CHECK IF THE ASSOCIATED MATRIX IS SQUARE.
 IF (SIZE(X).NE.A%N1 .OR. SIZE(Y).NE.A%N1) STOP "INCOMPATIBLE DIMENSIONS FOR THE BILINEAR&
												 &FORM COMPUTATION"
!INITIALIZE
 Q = 0

 DO i = 1, A%N1
 
	 DO j = A%IA1(i), A%IA1(i+1) - 1

		 k = A%JA1(j)

		 IF (i .EQ. k) Q = Q + A%A1(j)*X(i)*Y(k)

		 IF (i .LT. k) Q = Q + A%A1(j)*(X(i)*Y(k) + X(k)*Y(i))

	 ENDDO

 ENDDO
!***
! 
 END FUNCTION QFORM_IJA
!***
!
 END PROGRAM RDBLK_GS
!***
